from typing import List, Union
from numbers import Real

from qtpy import QtWidgets, QtGui
from qtpy.QtCore import QObject, Slot, Signal, Qt
import sys
import pyqtgraph

import pymodaq.utils.daq_utils as utils
from pymodaq.utils import data as data_mod
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.plotting.data_viewers.viewer import ViewerBase
from pymodaq.utils.managers.action_manager import ActionManager
from pymodaq.utils.plotting.widgets import PlotWidget
from pymodaq.utils.plotting.utils.plot_utils import Data0DWithHistory

import numpy as np
from collections import OrderedDict
import datetime

logger = set_logger(get_module_name(__file__))
PLOT_COLORS = utils.plot_colors


class DataDisplayer(QObject):
    """
    This Object deals with the display of 0D data  on a plotitem
    """

    updated_item = Signal(list)
    labels_changed = Signal(list)

    def __init__(self, plotitem: pyqtgraph.PlotItem):
        super().__init__()
        self._plotitem = plotitem
        self._plotitem.addLegend()
        self._plot_items: List[pyqtgraph.PlotDataItem] = []
        self._min_lines: List[pyqtgraph.InfiniteLine] = []
        self._max_lines: List[pyqtgraph.InfiniteLine] = []
        self._data = Data0DWithHistory()

        self._mins: List = []
        self._maxs: List = []

        self._show_lines: bool = False

        axis = self._plotitem.getAxis('bottom')
        axis.setLabel(text='Samples', units='S')

    @property
    def legend(self):
        return self._plotitem.legend

    @property
    def axis(self):
        return self._data.xaxis

    def clear_data(self):
        self._data.clear_data()
        self._mins = []
        self._maxs = []

    def update_axis(self, history_length: int):
        self._data.length = history_length

    def update_data(self, data: data_mod.DataRaw):
        if len(data) != len(self._plot_items):
            self.update_display_items(data)

        self._data.add_datas(data)
        for ind, data_str in enumerate(self._data.datas):
            self._plot_items[ind].setData(self._data.xaxis, self._data.datas[data_str])
        if len(self._mins) != len(self._data.datas):
            self._mins = []
            self._maxs = []

        for ind, label in enumerate(self._data.datas):
            if len(self._mins) != len(self._data.datas):
                self._mins.append(float(np.min(self._data.datas[label])))
                self._maxs.append(float(np.max(self._data.datas[label])))
            else:
                self._mins[ind] = min(self._mins[ind], float(np.min(self._data.datas[label])))
                self._maxs[ind] = max(self._maxs[ind], float(np.max(self._data.datas[label])))
            self._min_lines[ind].setValue(self._mins[ind])
            self._max_lines[ind].setValue(self._maxs[ind])

    def update_display_items(self, data: data_mod.DataRaw):
        while len(self._plot_items) > 0:
            plot_item = self._plotitem.removeItem(self._plot_items.pop(0))
            self.legend.removeItem(plot_item)
            self._plotitem.removeItem(self._max_lines.pop(0))
            self._plotitem.removeItem(self._min_lines.pop(0))

        for ind in range(len(data)):
            self._plot_items.append(pyqtgraph.PlotDataItem(pen=PLOT_COLORS[ind]))
            self._plotitem.addItem(self._plot_items[-1])
            self.legend.addItem(self._plot_items[-1], data.labels[ind])
            max_line = pyqtgraph.InfiniteLine(angle=0, pen=pyqtgraph.mkPen(color=PLOT_COLORS[ind], style=Qt.DashLine))
            min_line = pyqtgraph.InfiniteLine(angle=0, pen=pyqtgraph.mkPen(color=PLOT_COLORS[ind], style=Qt.DashLine))
            self._max_lines.append(max_line)
            self._min_lines.append(min_line)
            max_line.setVisible(self._show_lines)
            min_line.setVisible(self._show_lines)
            self._plotitem.addItem(self._max_lines[-1])
            self._plotitem.addItem(self._min_lines[-1])

        self.updated_item.emit(self._plot_items)
        self.labels_changed.emit(data.labels)

    def show_min_max(self, show=True):
        self._show_lines = show
        for line in self._max_lines:
            line.setVisible(show)
        for line in self._min_lines:
            line.setVisible(show)


class View0D(ActionManager, QObject):
    def __init__(self, parent_widget: QtWidgets.QWidget = None):
        QObject.__init__(self)
        ActionManager.__init__(self, toolbar=QtWidgets.QToolBar())

        self.data_displayer: DataDisplayer = None
        self.plot_widget: PlotWidget = PlotWidget()
        self.values_list = QtWidgets.QListWidget()

        self.setup_actions()

        self.parent_widget = parent_widget
        if self.parent_widget is None:
            self.parent_widget = QtWidgets.QWidget()
            self.parent_widget.show()

        self.data_displayer = DataDisplayer(self.plotitem)

        self._setup_widgets()
        self._connect_things()
        self._prepare_ui()

        self.get_action('Nhistory').setValue(200) #default history length

    def setup_actions(self):
        self.add_action('clear', 'Clear plot', 'clear2', 'Clear the current plots')
        self.add_widget('Nhistory', pyqtgraph.SpinBox, tip='Set the history length of the plot',
                        setters=dict(setMaximumWidth=100))
        self.add_action('show_data_as_list', 'Show numbers', 'ChnNum', 'If triggered, will display last data as numbers'
                                                                       'in a side panel', checkable=True)
        self.add_action('show_min_max', 'Show Min/Max lines', 'Statistics',
                        'If triggered, will display horizontal dashed lines for min/max of data', checkable=True)

    def _setup_widgets(self):
        self.parent_widget.setLayout(QtWidgets.QVBoxLayout())
        self.parent_widget.layout().addWidget(self.toolbar)

        splitter_hor = QtWidgets.QSplitter(Qt.Horizontal)
        self.parent_widget.layout().addWidget(splitter_hor)

        splitter_hor.addWidget(self.plot_widget)
        splitter_hor.addWidget(self.values_list)

        font = QtGui.QFont()
        font.setPointSize(20)
        self.values_list.setFont(font)

    def _connect_things(self):
        self.connect_action('clear', self.data_displayer.clear_data)
        self.connect_action('show_data_as_list', self.show_data_list)
        self.connect_action('Nhistory', self.data_displayer.update_axis, signal_name='valueChanged')
        self.connect_action('show_min_max', self.data_displayer.show_min_max)

    def _prepare_ui(self):
        """add here everything needed at startup"""
        self.values_list.setVisible(False)

    def get_double_clicked(self):
        return self.plot_widget.view.sig_double_clicked

    @property
    def plotitem(self):
        return self.plot_widget.plotItem

    def display_data(self, data: data_mod.DataRaw):
        self.data_displayer.update_data(data)
        if self.is_action_checked('show_data_as_list'):
            self.values_list.clear()
            self.values_list.addItems(['{:.03e}'.format(dat[0]) for dat in data])
            QtWidgets.QApplication.processEvents()

    def show_data_list(self, state=None):
        if state is None:
            state = self.is_action_checked('show_data_as_list')
        self.values_list.setVisible(state)


class Viewer0D(ViewerBase):
    """this plots 0D data on a plotwidget with history. Display as numbers in a table is possible.

    Datas and measurements are then exported with the signal data_to_export_signal
    """

    def __init__(self, parent=None, title=''):
        super().__init__(parent, title)
        self.view = View0D(self.parent)
        self._labels = []

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, labels):
        if labels != self._labels:
            self._labels = labels

    @Slot(list)
    def _show_data(self, data: data_mod.DataRaw):
        self.labels = data.labels
        self.view.display_data(data)
        self.data_to_export_signal.emit(self.data_to_export)


def main_view():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = View0D(widget)
    widget.show()
    sys.exit(app.exec_())


def main():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = Viewer0D(widget)
    from pymodaq.utils.daq_utils import gauss1D

    x = np.linspace(0, 200, 201)
    y1 = gauss1D(x, 75, 25) + 0.1*np.random.rand(len(x))
    y2 = 0.7 * gauss1D(x, 120, 50, 2) + 0.2*np.random.rand(len(x))
    widget.show()
    prog.get_action('show_data_as_list').trigger()
    for ind, data in enumerate(y1):
        prog.show_data(data_mod.DataRaw('mydata', data=[np.array([data]), np.array([y2[ind]])],
                                        labels=['lab1', 'lab2']))
        QtWidgets.QApplication.processEvents()

    sys.exit(app.exec_())


if __name__ == '__main__':  # pragma: no cover
    #main_view()
    main()



from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem, SimpleParameter
from pymodaq.utils.gui_utils.widgets import QLED


class LedParameterItem(WidgetParameterItem):
    """Registered parameter type which displays a QLineEdit"""

    def makeWidget(self):
        w = QLED()
        w.clickable = False
        w.set_as_false()
        w.sigChanged = w.value_changed
        w.value = w.get_state
        w.setValue = w.set_as
        self.hideWidget = False
        return w


class LedPushParameterItem(LedParameterItem):
    """Registered parameter type which displays a QLineEdit"""

    def makeWidget(self):
        w = QLED()
        w.clickable = True
        w.set_as_false()
        w.sigChanged = w.value_changed
        w.value = w.get_state
        w.setValue = w.set_as
        self.hideWidget = False
        return w


class LedParameter(SimpleParameter):
    itemClass = LedParameterItem

    def _interpretValue(self, v):
        return bool(v)


class LedPushParameter(SimpleParameter):
    itemClass = LedPushParameterItem

    def _interpretValue(self, v):
        return bool(v)



from pyqtgraph.parametertree.parameterTypes.basetypes import SimpleParameter, GroupParameter, GroupParameterItem
from .bool import BoolPushParameterItem
from .pixmap import PixmapParameterItem, PixmapCheckParameterItem
from .slide import SliderSpinBox, SliderParameterItem
from .led import LedPushParameter, LedParameter
from .date import DateParameter, DateTimeParameter, TimeParameter
from .list import ListParameter
from .table import TableParameter
from .tableview import TableViewParameter, TableViewCustom
from .itemselect import ItemSelectParameter
from .filedir import FileDirParameter
from .text import PlainTextPbParameter
from .numeric import NumericParameter

from pyqtgraph.parametertree.Parameter import registerParameterType, registerParameterItemType

registerParameterType('float', NumericParameter, override=True)
registerParameterType('int',   NumericParameter, override=True)
registerParameterItemType('bool_push', BoolPushParameterItem, SimpleParameter, override=True)
registerParameterItemType('pixmap', PixmapParameterItem, SimpleParameter, override=True)
registerParameterItemType('pixmap_check', PixmapCheckParameterItem, SimpleParameter, override=True)
registerParameterItemType('slide', SliderParameterItem, SimpleParameter, override=True)

registerParameterType('led', LedParameter, override=True)
registerParameterType('led_push', LedPushParameter, override=True)
registerParameterType('date', DateParameter, override=True)
registerParameterType('date_time', DateTimeParameter, override=True)
registerParameterType('time', TimeParameter, override=True)

registerParameterType('list', ListParameter, override=True)
registerParameterType('table', TableParameter, override=True)

registerParameterType('table_view', TableViewParameter, override=True)
registerParameterType('itemselect', ItemSelectParameter, override=True)
registerParameterType('browsepath', FileDirParameter, override=True)
registerParameterType('text_pb', PlainTextPbParameter, override=True)



# -*- coding: utf-8 -*-
"""
Created the 05/12/2022

@author: Sebastien Weber
"""
from typing import List, Tuple

import numpy as np

from qtpy import QtCore, QtWidgets
from pymodaq.utils.data import Axis, DataDistribution
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils import math_utils as mutils
from pymodaq.utils import config as configmod
from pymodaq.utils import gui_utils as gutils
from ..scan_factory import ScannerFactory, ScannerBase, ScanParameterManager
from pymodaq.utils.parameter import utils as putils
from pymodaq.utils.parameter.pymodaq_ptypes import TableViewCustom
from pymodaq.utils.plotting.scan_selector import Selector

logger = set_logger(get_module_name(__file__))
config = configmod.Config()


class TableModelSequential(gutils.TableModel):
    """Table Model for the Model/View Qt framework dedicated to the Sequential scan mode"""
    def __init__(self, data, **kwargs):
        header = ['Actuator', 'Start', 'Stop', 'Step']
        if 'header' in kwargs:
            header = kwargs.pop('header')
        editable = [False, True, True, True]
        if 'editable' in kwargs:
            editable = kwargs.pop('editable')
        super().__init__(data, header, editable=editable, **kwargs)

    def __repr__(self):
        return f'{self.__class__.__name__} from module {self.__class__.__module__}'

    def validate_data(self, row, col, value):
        """
        make sure the values and signs of the start, stop and step values are "correct"
        Parameters
        ----------
        row: (int) row within the table that is to be changed
        col: (int) col within the table that is to be changed
        value: (float) new value for the value defined by row and col

        Returns
        -------
        bool: True is the new value is fine (change some other values if needed) otherwise False
        """
        start = self.data(self.index(row, 1), QtCore.Qt.DisplayRole)
        stop = self.data(self.index(row, 2), QtCore.Qt.DisplayRole)
        step = self.data(self.index(row, 3), QtCore.Qt.DisplayRole)
        isstep = False
        if col == 1:  # the start
            start = value
        elif col == 2:  # the stop
            stop = value
        elif col == 3:  # the step
            isstep = True
            step = value

        if np.abs(step) < 1e-12 or start == stop:
            return False
        if np.sign(stop - start) != np.sign(step):
            if isstep:
                self._data[row][2] = -stop
            else:
                self._data[row][3] = -step
        return True


@ScannerFactory.register()
class SequentialScanner(ScannerBase):
    scan_type = 'Sequential'
    scan_subtype = 'Linear'
    save_settings = False  # not easy to save table content in a toml...
    params = [
        {'title': 'Sequences', 'name': 'seq_table', 'type': 'table_view',
         'delegate': gutils.SpinBoxDelegate},
              ]
    distribution = DataDistribution['uniform']
    n_axes = 1

    def __init__(self, actuators: List[str]):

        self.table_model: TableModelSequential = None
        self.table_view: TableViewCustom = None
        super().__init__(actuators)
        self.update_model()

    @property
    def actuators(self):
        return self._actuators

    @actuators.setter
    def actuators(self, actuators_name):
        self._actuators = actuators_name
        self.update_model()

    def update_model(self, init_data=None):
        if init_data is None:
            if self.table_model is not None:
                init_data = []
                names = [row[0] for row in self.table_model.get_data_all()]
                for act in self._actuators:
                    if act.title in names:
                        ind_row = names.index(act.title)
                        init_data.append(self.table_model.get_data_all()[ind_row])
                    else:
                        init_data.append([act.title, 0., 1., 0.1])
            else:
                init_data = [[act.title, 0., 1., 0.1] for act in self._actuators]
        self.table_model = TableModelSequential(init_data, )
        self.table_view = putils.get_widget_from_tree(self.settings_tree, TableViewCustom)[0]
        self.settings.child('seq_table').setValue(self.table_model)
        self.n_axes = len(self._actuators)
        self.update_table_view()

    def get_pos(self):
        starts = np.array([self.table_model.get_data(ind, 1)
                           for ind in range(self.table_model.rowCount(None))])
        stops = np.array([self.table_model.get_data(ind, 2)
                          for ind in range(self.table_model.rowCount(None))])
        steps = np.array([self.table_model.get_data(ind, 3)
                          for ind in range(self.table_model.rowCount(None))])
        return starts, stops, steps

    def evaluate_steps(self) -> int:
        starts, stops, steps = self.get_pos()
        n_steps = 1
        for ind in range(starts.size):
            n_steps *= np.abs((stops[ind] - starts[ind]) / steps[ind]) + 1
        return int(n_steps)

    @staticmethod
    def pos_above_stops(positions, steps, stops):
        state = []
        for pos, step, stop in zip(positions, steps, stops):
            if step >= 0:
                state.append(pos > stop)
            else:
                state.append(pos < stop)
        return state

    def update_table_view(self):
        self.table_view.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.table_view.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        styledItemDelegate = QtWidgets.QStyledItemDelegate()
        styledItemDelegate.setItemEditorFactory(gutils.SpinBoxDelegate())
        self.table_view.setItemDelegate(styledItemDelegate)

        self.table_view.setDragEnabled(True)
        self.table_view.setDropIndicatorShown(True)
        self.table_view.setAcceptDrops(True)
        self.table_view.viewport().setAcceptDrops(True)
        self.table_view.setDefaultDropAction(QtCore.Qt.MoveAction)
        self.table_view.setDragDropMode(QtWidgets.QTableView.InternalMove)
        self.table_view.setDragDropOverwriteMode(False)

    def set_scan(self):
        starts, stops, steps = self.get_pos()
        all_positions = [starts.copy()]
        positions = starts.copy()
        state = self.pos_above_stops(positions, steps, stops)
        if len(state) != 0:
            while not state[0]:
                if not np.any(np.array(state)):
                    positions[-1] += steps[-1]

                else:
                    indexes_true = np.where(np.array(state))
                    positions[indexes_true[-1][0]] = starts[indexes_true[-1][0]]
                    positions[indexes_true[-1][0] - 1] += steps[indexes_true[-1][0] - 1]

                state = self.pos_above_stops(positions, steps, stops)
                if not np.any(np.array(state)):
                    all_positions.append(positions.copy())

        self.get_info_from_positions(np.array(all_positions))

    def get_nav_axes(self) -> List[Axis]:
        return [Axis(label=f'{act.title}', units=act.units, data=self.axes_unique[ind], index=ind)
                for ind, act in enumerate(self.actuators)]

    def get_indexes_from_scan_index(self, scan_index: int) -> Tuple[int]:
        """To be reimplemented. Calculations of indexes within the scan"""
        return tuple(self.axes_indexes[scan_index])

    def get_scan_shape(self) -> Tuple[int]:
        return tuple([len(axis) for axis in self.axes_unique])

    def update_from_scan_selector(self, scan_selector: Selector):
        coordinates = scan_selector.get_coordinates()
        pass


from qtpy import QtWidgets
import sys
import os

from pymodaq.utils.parameter import ioxml
from pymodaq.utils.parameter import ParameterTree, Parameter
from pyqtgraph.parametertree.Parameter import registerParameterType
from pyqtgraph.parametertree.parameterTypes.basetypes import GroupParameter
from pymodaq.utils.gui_utils import select_file

# check if overshoot_configurations directory exists on the drive
from pymodaq.utils.config import get_set_overshoot_path

overshoot_path = get_set_overshoot_path()


class PresetScalableGroupMove(GroupParameter):
    """
        |

        ================ =============
        **Attributes**    **Type**
        *opts*            dictionnary
        ================ =============

        See Also
        --------
        hardware.DAQ_Move_Stage_type
    """

    def __init__(self, **opts):
        opts['type'] = 'groupmoveover'
        opts['addText'] = "Add"
        opts['addList'] = opts['movelist']
        super().__init__(**opts)

    def addNew(self, name):
        """
            Add a child.

            =============== ===========
            **Parameters**   **Type**
            *typ*            string
            =============== ===========
        """
        name_prefix = 'move'
        child_indexes = [int(par.name()[len(name_prefix) + 1:]) for par in self.children()]
        if not child_indexes:
            newindex = 0
        else:
            newindex = max(child_indexes) + 1

        child = {'title': name, 'name': f'{name_prefix}{newindex:02.0f}', 'type': 'group', 'removable': True,
                 'children': [
                     {'title': 'Move if overshoot?:', 'name': 'move_overshoot', 'type': 'bool', 'value': True},
                     {'title': 'Position:', 'name': 'position', 'type': 'float', 'value': 0}], 'removable': True,
                 'renamable': False}

        self.addChild(child)


registerParameterType('groupmoveover', PresetScalableGroupMove, override=True)


class PresetScalableGroupDet(GroupParameter):
    """
        =============== ==============
        **Attributes**    **Type**
        *opts*            dictionnary
        *options*         string list
        =============== ==============

        See Also
        --------
    """

    def __init__(self, **opts):
        opts['type'] = 'groupdetover'
        opts['addText'] = "Add"
        opts['addList'] = opts['detlist']
        opts['movelist'] = opts['movelist']

        super().__init__(**opts)

    def addNew(self, name):
        """
            Add a child.

            =============== ===========  ================
            **Parameters**    **Type**   **Description*
            *typ*             string     the viewer name
            =============== ===========  ================
        """
        try:
            name_prefix = 'det'
            child_indexes = [int(par.name()[len(name_prefix) + 1:]) for par in self.children()]
            if not child_indexes:
                newindex = 0
            else:
                newindex = max(child_indexes) + 1

            child = {'title': name, 'name': f'{name_prefix}{newindex:02.0f}', 'type': 'group', 'children': [
                {'title': 'Trig overshoot?:', 'name': 'trig_overshoot', 'type': 'bool', 'value': True},
                {'title': 'Overshoot value:', 'name': 'overshoot_value', 'type': 'float', 'value': 20},
                {'title': 'Triggered Moves:', 'name': 'params', 'type': 'groupmoveover',
                 'movelist': self.opts['movelist']}], 'removable': True, 'renamable': False}

            self.addChild(child)
        except Exception as e:
            print(str(e))


registerParameterType('groupdetover', PresetScalableGroupDet, override=True)


class OvershootManager:
    def __init__(self, msgbox=False, det_modules=[], actuators_modules=[]):

        self.overshoot_params = None
        self.det_modules = det_modules
        self.actuators_modules = actuators_modules

        if msgbox:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText("Overshoot Manager?")
            msgBox.setInformativeText("What do you want to do?")
            cancel_button = msgBox.addButton(QtWidgets.QMessageBox.Cancel)
            new_button = msgBox.addButton("New", QtWidgets.QMessageBox.ActionRole)
            modify_button = msgBox.addButton('Modify', QtWidgets.QMessageBox.AcceptRole)
            msgBox.setDefaultButton(QtWidgets.QMessageBox.Cancel)
            ret = msgBox.exec()

            if msgBox.clickedButton() == new_button:
                self.set_new_overshoot()

            elif msgBox.clickedButton() == modify_button:
                path = select_file(start_path=overshoot_path, save=False, ext='xml')
                if path != '':
                    self.set_file_overshoot(str(path))
            else:  # cancel
                pass

    def set_file_overshoot(self, filename, show=True):
        """

        """
        children = ioxml.XML_file_to_parameter(filename)
        self.overshoot_params = Parameter.create(title='Overshoot', name='Overshoot', type='group', children=children)
        if show:
            self.show_overshoot()

    def set_new_overshoot(self, file=None):
        if file is None:
            file = 'overshoot_default'
        param = [{'title': 'Filename:', 'name': 'filename', 'type': 'str', 'value': file}]
        params_det = [{'title': 'Detectors:', 'name': 'Detectors', 'type': 'groupdetover', 'detlist': self.det_modules,
                       'movelist': self.actuators_modules}]  # [PresetScalableGroupDet(name="Detectors")]
        self.overshoot_params = Parameter.create(title='Preset', name='Preset', type='group',
                                                 children=param + params_det)

        self.show_overshoot()

    def show_overshoot(self):
        """

        """
        dialog = QtWidgets.QDialog()
        vlayout = QtWidgets.QVBoxLayout()
        tree = ParameterTree()
        tree.setMinimumWidth(400)
        tree.setMinimumHeight(500)
        tree.setParameters(self.overshoot_params, showTop=False)

        vlayout.addWidget(tree)
        dialog.setLayout(vlayout)
        buttonBox = QtWidgets.QDialogButtonBox(parent=dialog)

        buttonBox.addButton('Save', buttonBox.AcceptRole)
        buttonBox.accepted.connect(dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(dialog.reject)

        vlayout.addWidget(buttonBox)
        dialog.setWindowTitle('Fill in information about this managers')
        res = dialog.exec()

        if res == dialog.Accepted:
            # save managers parameters in a xml file
            # start = os.path.split(os.path.split(os.path.realpath(__file__))[0])[0]
            # start = os.path.join("..",'daq_scan')
            ioxml.parameter_to_xml_file(
                self.overshoot_params, os.path.join(overshoot_path, self.overshoot_params.child('filename').value()))


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    prog = OvershootManager(True, ['det camera', 'det current'], ['Move X', 'Move Y'])

    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Created the 28/10/2022

@author: Sebastien Weber
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod, abstractproperty
import numbers
import numpy as np
from typing import List, Tuple, Union, Any
from typing import Iterable as IterableType
from collections.abc import Iterable
import logging

import warnings
from time import time
import copy

from multipledispatch import dispatch
from pymodaq.utils.enums import BaseEnum, enum_checker
from pymodaq.utils.messenger import deprecation_msg
from pymodaq.utils.daq_utils import find_objects_in_list_from_attr_name_val
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.slicing import SpecialSlicersData
from pymodaq.utils import math_utils as mutils

logger = set_logger(get_module_name(__file__))


class DataIndexWarning(Warning):
    pass


class DataTypeWarning(Warning):
    pass


class DataDimWarning(Warning):
    pass


class DataSizeWarning(Warning):
    pass


WARNINGS = [DataIndexWarning, DataTypeWarning, DataDimWarning, DataSizeWarning]

if logging.getLevelName(logger.level) == 'DEBUG':
    for warning in WARNINGS:
        warnings.filterwarnings('default', category=warning)
else:
    for warning in WARNINGS:
        warnings.filterwarnings('ignore', category=warning)


class DataShapeError(Exception):
    pass


class DataLengthError(Exception):
    pass


class DataDimError(Exception):
    pass


class DwaType(BaseEnum):
    DataWithAxes = 0
    DataRaw = 1
    DataActuator = 2
    DataFromPlugins = 3
    DataCalculated = 4


class DataDim(BaseEnum):
    """Enum for dimensionality representation of data"""
    Data0D = 0
    Data1D = 1
    Data2D = 2
    DataND = 3

    def __le__(self, other_dim: 'DataDim'):
        return self.value.__le__(other_dim.value)

    def __lt__(self, other_dim: 'DataDim'):
        return self.value.__lt__(other_dim.value)

    def __ge__(self, other_dim: 'DataDim'):
        other_dim = enum_checker(DataDim, other_dim)
        return self.value.__ge__(other_dim.value)

    def __gt__(self, other_dim: 'DataDim'):
        return self.value.__gt__(other_dim.value)

    @property
    def dim_index(self):
       return self.value


class DataSource(BaseEnum):
    """Enum for source of data"""
    raw = 0
    calculated = 1


class DataDistribution(BaseEnum):
    """Enum for distribution of data"""
    uniform = 0
    spread = 1


class Axis:
    """Object holding info and data about physical axis of some data

    In case the axis's data is linear, store the info as a scale and offset else store the data

    Parameters
    ----------
    label: str
        The label of the axis, for instance 'time' for a temporal axis
    units: str
        The units of the data in the object, for instance 's' for seconds
    data: ndarray
        A 1D ndarray holding the data of the axis
    index: int
        an integer representing the index of the Data object this axis is related to
    scaling: float
        The scaling to apply to a linspace version in order to obtain the proper scaling
    offset: float
        The offset to apply to a linspace/scaled version in order to obtain the proper axis
    spread_order: int
        An integer needed in the case where data has a spread DataDistribution. It refers to the index along the data's
        spread_index dimension

    Examples
    --------
    >>> axis = Axis('myaxis', units='seconds', data=np.array([1,2,3,4,5]), index=0)
    """

    def __init__(self, label: str = '', units: str = '', data: np.ndarray = None, index: int = 0, scaling=None,
                 offset=None, spread_order: int = 0):
        super().__init__()

        self.iaxis: Axis = SpecialSlicersData(self, False)

        self._size = None
        self._data = None
        self._index = None
        self._label = None
        self._units = None
        self._scaling = scaling
        self._offset = offset

        self.units = units
        self.label = label
        self.data = data
        self.index = index
        self.spread_order = spread_order
        if (scaling is None or offset is None) and data is not None:
            self.get_scale_offset_from_data(data)

    def copy(self):
        return copy.copy(self)

    @property
    def label(self) -> str:
        """str: get/set the label of this axis"""
        return self._label

    @label.setter
    def label(self, lab: str):
        if not isinstance(lab, str):
            raise TypeError('label for the Axis class should be a string')
        self._label = lab

    @property
    def units(self) -> str:
        """str: get/set the units for this axis"""
        return self._units

    @units.setter
    def units(self, units: str):
        if not isinstance(units, str):
            raise TypeError('units for the Axis class should be a string')
        self._units = units

    @property
    def index(self) -> int:
        """int: get/set the index this axis corresponds to in a DataWithAxis object"""
        return self._index

    @index.setter
    def index(self, ind: int):
        self._check_index_valid(ind)
        self._index = ind

    @property
    def data(self):
        """np.ndarray: get/set the data of Axis"""
        return self._data

    @data.setter
    def data(self, data: np.ndarray):
        if data is not None:
            self._check_data_valid(data)
            self.get_scale_offset_from_data(data)
            self._size = data.size
        else:
            self._size = 0
        self._data = data

    def get_data(self) -> np.ndarray:
        """Convenience method to obtain the axis data (usually None because scaling and offset are used)"""
        return self._data if self._data is not None else self._linear_data(self.size)

    def get_scale_offset_from_data(self, data: np.ndarray = None):
        """Get the scaling and offset from the axis's data

        If data is not None, extract the scaling and offset

        Parameters
        ----------
        data: ndarray
        """
        if data is None and self._data is not None:
            data = self._data

        if self.is_axis_linear(data):
            if len(data) == 1:
                self._scaling = 1
            else:
                self._scaling = np.mean(np.diff(data))
            self._offset = data[0]
            self._data = None

    def is_axis_linear(self, data=None):
        if data is None:
            data = self.get_data()
        if data is not None:
            return np.allclose(np.diff(data), np.mean(np.diff(data)))
        else:
            return False

    @property
    def scaling(self):
        return self._scaling

    @scaling.setter
    def scaling(self, _scaling: float):
        self._scaling = _scaling

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, _offset: float):
        self._offset = _offset

    @property
    def size(self) -> int:
        """int: get/set the size/length of the 1D ndarray"""
        return self._size

    @size.setter
    def size(self, _size: int):
        if self._data is None:
            self._size = _size

    @staticmethod
    def _check_index_valid(index: int):
        if not isinstance(index, int):
            raise TypeError('index for the Axis class should be a positive integer')
        elif index < 0:
            raise ValueError('index for the Axis class should be a positive integer')

    @staticmethod
    def _check_data_valid(data):
        if not isinstance(data, np.ndarray):
            raise TypeError(f'data for the Axis class should be a 1D numpy array')
        elif len(data.shape) != 1:
            raise ValueError(f'data for the Axis class should be a 1D numpy array')

    def _linear_data(self, nsteps: int):
        """create axis data with a linear version using scaling and offset"""
        return self._offset + self._scaling * np.linspace(0, nsteps-1, nsteps)

    def create_linear_data(self, nsteps:int):
        """replace the axis data with a linear version using scaling and offset"""
        self.data = self._linear_data(nsteps)

    @staticmethod
    def create_simple_linear_data(nsteps: int):
        return np.linspace(0, nsteps-1, nsteps)

    def __len__(self):
        return self.size

    def _compute_slices(self, slices, *ignored, **ignored_also):
        return slices

    def _slicer(self, _slice, *ignored, **ignored_also):
        ax: Axis = copy.deepcopy(self)
        if isinstance(_slice, int):
            ax.data = np.array([ax.get_data()[_slice]])
            return ax
        elif _slice is Ellipsis:
            return ax
        elif isinstance(_slice, slice):
            if ax._data is not None:
                ax.data = ax._data.__getitem__(_slice)
                return ax
            else:
                start = _slice.start if _slice.start is not None else 0
                stop = _slice.stop if _slice.stop is not None else self.size

                ax._offset = ax.offset + start * ax.scaling
                ax._size = stop - start
                return ax

    def __getitem__(self, item):
        if hasattr(self, item):
            # for when axis was a dict
            deprecation_msg('attributes from an Axis object should not be fetched using __getitem__')
            return getattr(self, item)

    def __repr__(self):
        return f'{self.__class__.__name__}: <label: {self.label}> - <units: {self.units}> - <index: {self.index}>'

    def __mul__(self, scale: numbers.Real):
        if isinstance(scale, numbers.Real):
            ax = copy.deepcopy(self)
            if self.data is not None:
                ax.data *= scale
            else:
                ax._offset *= scale
                ax._scaling *= scale
            return ax

    def __add__(self, offset: numbers.Real):
        if isinstance(offset, numbers.Real):
            ax = copy.deepcopy(self)
            if self.data is not None:
                ax.data += offset
            else:
                ax._offset += offset
            return ax

    def __eq__(self, other):
        eq = self.label == other.label
        eq = eq and (self.units == other.units)
        eq = eq and (self.index == other.index)
        if self.data is not None and other.data is not None:
            eq = eq and (np.allclose(self.data, other.data))
        else:
            eq = eq and self.offset == other.offset
            eq = eq and self.scaling == other.scaling

        return eq

    def mean(self):
        if self._data is not None:
            return np.mean(self._data)
        else:
            return self.offset + self.size / 2 * self.scaling

    def min(self):
        if self._data is not None:
            return np.min(self._data)
        else:
            return self.offset + (self.size * self.scaling if self.scaling < 0 else 0)

    def max(self):
        if self._data is not None:
            return np.max(self._data)
        else:
            return self.offset + (self.size * self.scaling if self.scaling > 0 else 0)

    def find_index(self, threshold: float) -> int:
        """find the index of the threshold value within the axis"""
        if threshold < self.min():
            return 0
        elif threshold > self.max():
            return len(self) - 1
        elif self._data is not None:
            return mutils.find_index(self._data, threshold)[0][0]
        else:
            return int((threshold - self.offset) / self.scaling)

    def find_indexes(self, thresholds: IterableType[float]) -> IterableType[int]:
        return [self.find_index(threshold) for threshold in thresholds]


class NavAxis(Axis):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        deprecation_msg('NavAxis should not be used anymore, please use Axis object with correct index.'
                        'The navigation index should be specified in the Data object')


class DataLowLevel:
    """Abstract object for all Data Object

    Parameters
    ----------
    name: str
        the identifier of the data

    Attributes
    ----------
    name: str
    timestamp: float
        Time in seconds since epoch. See method time.time()
    """

    def __init__(self, name: str):
        self._timestamp = time()
        self._name = name

    @property
    def name(self):
        """Get/Set the identifier of the data"""
        return self._name

    @name.setter
    def name(self, other_name: str):
        self._name = other_name

    @property
    def timestamp(self):
        """Get/Set the timestamp of when the object has been created"""
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: float):
        """The timestamp of when the object has been created"""
        self._timestamp = timestamp


class DataBase(DataLowLevel):
    """Base object to store homogeneous data and metadata generated by pymodaq's objects. To be inherited for real data

    Parameters
    ----------
    name: str
        the identifier of these data
    source: DataSource or str
        Enum specifying if data are raw or processed (for instance from roi)
    dim: DataDim or str
        The identifier of the data type
    distribution: DataDistribution or str
        The distribution type of the data: uniform if distributed on a regular grid or spread if on specific
        unordered points
    data: list of ndarray
        The data the object is storing
    labels: list of str
        The labels of the data nd-arrays
    origin: str
        An identifier of the element where the data originated, for instance the DAQ_Viewer's name. Used when appending
        DataToExport in DAQ_Scan to disintricate from which origin data comes from when scanning multiple detectors.
    kwargs: named parameters
        All other parameters are stored dynamically using the name/value pair. The name of these extra parameters are
        added into the extra_attributes attribute

    Attributes
    ----------
    name: str
        the identifier of these data
    source: DataSource or str
        Enum specifying if data are raw or processed (for instance from roi)
    dim: DataDim or str
        The identifier of the data type
    distribution: DataDistribution or str
        The distribution type of the data: uniform if distributed on a regular grid or spread if on specific
        unordered points
    data: list of ndarray
        The data the object is storing
    labels: list of str
        The labels of the data nd-arrays
    origin: str
        An identifier of the element where the data originated, for instance the DAQ_Viewer's name. Used when appending
        DataToExport in DAQ_Scan to disintricate from which origin data comes from when scanning multiple detectors.
    shape: Tuple[int]
        The shape of the underlying data
    size: int
        The size of the ndarrays stored in the object
    length: int
        The number of ndarrays stored in the object
    extra_attributes: List[str]
        list of string giving identifiers of the attributes added dynamically at the initialization (for instance
        to save extra metadata using the DataSaverLoader

    See Also
    --------
    DataWithAxes, DataFromPlugins, DataRaw, DataSaverLoader

    Examples
    --------
    >>> import numpy as np
    >>> from pymodaq.utils.data import DataBase, DataSource, DataDim, DataDistribution
    >>> data = DataBase('mydata', source=DataSource['raw'], dim=DataDim['Data1D'], \
    distribution=DataDistribution['uniform'], data=[np.array([1.,2.,3.]), np.array([4.,5.,6.])],\
    labels=['channel1', 'channel2'], origin='docutils code')
    >>> data.dim
    <DataDim.Data1D: 1>
    >>> data.source
    <DataSource.raw: 0>
    >>> data.shape
    (3,)
    >>> data.length
    2
    >>> data.size
    3
    """

    def __init__(self, name: str, source: DataSource = None, dim: DataDim = None,
                 distribution: DataDistribution = DataDistribution['uniform'], data: List[np.ndarray] = None,
                 labels: List[str] = [], origin: str = '', **kwargs):

        super().__init__(name=name)
        self._iter_index = 0
        self._shape = None
        self._size = None
        self._data = None
        self._length = None
        self._labels = None
        self._dim = dim
        self.origin = origin

        source = enum_checker(DataSource, source)
        self._source = source

        distribution = enum_checker(DataDistribution, distribution)
        self._distribution = distribution

        self.data = data  # dim consistency is actually checked within the setter method

        self._check_labels(labels)
        self.extra_attributes = []
        self.add_extra_attribute(**kwargs)

    def add_extra_attribute(self, **kwargs):
        for key in kwargs:
            if key not in self.extra_attributes:
                self.extra_attributes.append(key)
            setattr(self, key, kwargs[key])

    def get_full_name(self) -> str:
        """Get the data ful name including the origin attribute into the returned value

        Returns
        -------
        str: the name of the ataWithAxes data constructed as : origin/name

        Examples
        --------
        d0 = DataBase(name='datafromdet0', origin='det0')
        """
        return f'{self.origin}/{self.name}'

    def __repr__(self):
        return f'{self.__class__.__name__} <{self.name}> <{self.dim}> <{self.source}> <{self.shape}>'

    def __len__(self):
        return self.length

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        if self._iter_index < len(self):
            self._iter_index += 1
            return self.data[self._iter_index-1]
        else:
            raise StopIteration

    def __getitem__(self, item) -> np.ndarray:
        if isinstance(item, int) and item < len(self):
            return self.data[item]
        else:
            raise IndexError(f'The index should be an integer lower than the data length')

    def __setitem__(self, key, value):
        if isinstance(key, int) and key < len(self) and isinstance(value, np.ndarray) and value.shape == self.shape:
            self.data[key] = value
        else:
            raise IndexError(f'The index should be an positive integer lower than the data length')

    def __add__(self, other: object):
        if isinstance(other, DataBase) and len(other) == len(self):
            new_data = copy.deepcopy(self)
            for ind_array in range(len(new_data)):
                if self[ind_array].shape != other[ind_array].shape:
                    raise ValueError('The shapes of arrays stored into the data are not consistent')
                new_data[ind_array] = self[ind_array] + other[ind_array]
            return new_data
        elif isinstance(other, numbers.Number) and self.length == 1 and self.size == 1:
            new_data = copy.deepcopy(self)
            new_data = new_data + DataActuator(data=other)
            return new_data
        else:
            raise TypeError(f'Could not add a {other.__class__.__name__} or a {self.__class__.__name__} '
                            f'of a different length')

    def __sub__(self, other: object):
        if isinstance(other, DataBase) and len(other) == len(self):
            new_data = copy.deepcopy(self)
            for ind_array in range(len(new_data)):
                new_data[ind_array] = self[ind_array] - other[ind_array]
            return new_data
        elif isinstance(other, numbers.Number) and self.length == 1 and self.size == 1:
            new_data = copy.deepcopy(self)
            new_data = new_data - DataActuator(data=other)
            return new_data
        else:
            raise TypeError(f'Could not substract a {other.__class__.__name__} or a {self.__class__.__name__} '
                            f'of a different length')

    def __mul__(self, other):
        if isinstance(other, numbers.Number):
            new_data = copy.deepcopy(self)
            for ind_array in range(len(new_data)):
                new_data[ind_array] = self[ind_array] * other
            return new_data
        else:
            raise TypeError(f'Could not multiply a {other.__class__.__name__} and a {self.__class__.__name__} '
                            f'of a different length')

    def __truediv__(self, other):
        if isinstance(other, numbers.Number):
            return self * (1 / other)
        else:
            raise TypeError(f'Could not divide a {other.__class__.__name__} and a {self.__class__.__name__} '
                            f'of a different length')

    def _comparison_common(self, other, operator='__eq__'):
        if isinstance(other, DataBase):
            if not(self.name == other.name and len(self) == len(other)):
                return False
            eq = True
            for ind in range(len(self)):
                if self[ind].shape != other[ind].shape:
                    eq = False
                    break
                eq = eq and np.all(getattr(self[ind], operator)(other[ind]))
            return eq
        elif isinstance(other, numbers.Number):
            return np.all(getattr(self[0], operator)(other))
        else:
            raise TypeError()

    def __eq__(self, other):
        return self._comparison_common(other, '__eq__')

    def __le__(self, other):
        return self._comparison_common(other, '__le__')

    def __lt__(self, other):
        return self._comparison_common(other, '__lt__')

    def __ge__(self, other):
        return self._comparison_common(other, '__ge__')

    def __gt__(self, other):
        return self._comparison_common(other, '__gt__')

    def average(self, other: 'DataBase', weight: int) -> 'DataBase':
        """ Compute the weighted average between self and other DataBase

        Parameters
        ----------
        other_data: DataBase
        weight: int
            The weight the 'other' holds with respect to self
        Returns
        -------
        DataBase: the averaged DataBase object
        """
        if isinstance(other, DataBase) and len(other) == len(self) and isinstance(weight, numbers.Number):
            return (other * weight + self) / (weight + 1)
        else:
            raise TypeError(f'Could not average a {other.__class__.__name__} or a {self.__class__.__name__} '
                            f'of a different length')

    def abs(self):
        """ Take the absolute value of itself"""
        new_data = copy.copy(self)
        new_data.data = [np.abs(dat) for dat in new_data]
        return new_data

    def flipud(self):
        """Reverse the order of elements along axis 0 (up/down)"""
        new_data = copy.copy(self)
        new_data.data = [np.flipud(dat) for dat in new_data]
        return new_data

    def fliplr(self):
        """Reverse the order of elements along axis 1 (left/right)"""
        new_data = copy.copy(self)
        new_data.data = [np.fliplr(dat) for dat in new_data]
        return new_data

    def append(self, data: DataWithAxes):
        for dat in data:
            if dat.shape != self.shape:
                raise DataShapeError('Cannot append those ndarrays, they don\'t have the same shape as self')
        self.data = self.data + data.data
        self.labels.extend(data.labels)

    @property
    def shape(self):
        """The shape of the nd-arrays"""
        return self._shape

    @property
    def size(self):
        """The size of the nd-arrays"""
        return self._size

    @property
    def dim(self):
        """DataDim: the enum representing the dimensionality of the stored data"""
        return self._dim

    def set_dim(self, dim: Union[DataDim, str]):
        """Addhoc modification of dim independantly of the real data shape, should be used with extra care"""
        self._dim = enum_checker(DataDim, dim)

    @property
    def source(self):
        """DataSource: the enum representing the source of the data"""
        return self._source

    @source.setter
    def source(self, source_type: Union[str, DataSource]):
        """DataSource: the enum representing the source of the data"""
        source_type = enum_checker(DataSource, source_type)
        self._source = source_type

    @property
    def distribution(self):
        """DataDistribution: the enum representing the distribution of the stored data"""
        return self._distribution

    @property
    def length(self):
        """The length of data. This is the length of the list containing the nd-arrays"""
        return self._length

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, labels: List['str']):
        self._check_labels(labels)

    def _check_labels(self, labels: List['str']):
        if labels is None:
            labels = []
        else:
            labels = labels[:]
        while len(labels) < self.length:
            labels.append(f'CH{len(labels):02d}')
        self._labels = labels

    def get_data_index(self, index: int = 0):
        """Get the data by its index in the list"""
        return self.data[index]

    @staticmethod
    def _check_data_type(data: List[np.ndarray]) -> List[np.ndarray]:
        """make sure data is a list of nd-arrays"""
        is_valid = True
        if data is None:
            is_valid = False
        if not isinstance(data, list):
            # try to transform the data to regular type
            if isinstance(data, np.ndarray):
                warnings.warn(DataTypeWarning(f'Your data should be a list of numpy arrays not just a single numpy'
                                              f' array, wrapping them with a list'))
                data = [data]
            elif isinstance(data, numbers.Number):
                warnings.warn(DataTypeWarning(f'Your data should be a list of numpy arrays not just a single numpy'
                                              f' array, wrapping them with a list'))
                data = [np.array([data])]
            else:
                is_valid = False
        if isinstance(data, list):
            if len(data) == 0:
                is_valid = False
            if not isinstance(data[0], np.ndarray):
                is_valid = False
            elif len(data[0].shape) == 0:
                is_valid = False
        if not is_valid:
            raise TypeError(f'Data should be an non-empty list of non-empty numpy arrays')
        return data

    def check_shape_from_data(self, data: List[np.ndarray]):
        self._shape = data[0].shape

    @staticmethod
    def _get_dim_from_data(data: List[np.ndarray]) -> DataDim:
        shape = data[0].shape
        size = data[0].size
        if len(shape) == 1 and size == 1:
            dim = DataDim['Data0D']
        elif len(shape) == 1 and size > 1:
            dim = DataDim['Data1D']
        elif len(shape) == 2:
            dim = DataDim['Data2D']
        else:
            dim = DataDim['DataND']
        return dim

    def get_dim_from_data(self, data: List[np.ndarray]):
        """Get the dimensionality DataDim from data"""
        self.check_shape_from_data(data)
        self._size = data[0].size
        self._length = len(data)
        if len(self._shape) == 1 and self._size == 1:
            dim = DataDim['Data0D']
        elif len(self._shape) == 1 and self._size > 1:
            dim = DataDim['Data1D']
        elif len(self._shape) == 2:
            dim = DataDim['Data2D']
        else:
            dim = DataDim['DataND']
        return dim

    def _check_shape_dim_consistency(self, data: List[np.ndarray]):
        """Process the dim from data or make sure data and DataDim are coherent"""
        dim = self.get_dim_from_data(data)
        if self._dim is None:
            self._dim = dim
        else:
            self._dim = enum_checker(DataDim, self._dim)
            if self._dim != dim:
                warnings.warn(DataDimWarning('The specified dimensionality is not coherent with the data shape, '
                                             'replacing it'))
                self._dim = dim

    def _check_same_shape(self, data: List[np.ndarray]):
        """Check that all nd-arrays have the same shape"""
        for dat in data:
            if dat.shape != self.shape:
                raise DataShapeError('The shape of the ndarrays in data is not the same')

    @property
    def data(self) -> List[np.ndarray]:
        """List[np.ndarray]: get/set (and check) the data the object is storing"""
        return self._data

    @data.setter
    def data(self, data: List[np.ndarray]):
        data = self._check_data_type(data)
        self._check_shape_dim_consistency(data)
        self._check_same_shape(data)
        self._data = data


class AxesManagerBase:
    def __init__(self, data_shape: Tuple[int], axes: List[Axis], nav_indexes=None, sig_indexes=None, **kwargs):
        self._data_shape = data_shape[:]  # initial shape needed for self._check_axis
        self._axes = axes[:]
        self._nav_indexes = nav_indexes
        self._sig_indexes = sig_indexes if sig_indexes is not None else self.compute_sig_indexes()

        self._check_axis(self._axes)
        self._manage_named_axes(self._axes, **kwargs)

    @property
    def axes(self):
        return self._axes

    @axes.setter
    def axes(self, axes: List[Axis]):
        self._axes = axes[:]
        self._check_axis(self._axes)

    @abstractmethod
    def _check_axis(self, axes):
        ...

    @abstractmethod
    def get_sorted_index(self, axis_index: int = 0, spread_index=0) -> Tuple[np.ndarray, Tuple[slice]]:
        """ Get the index to sort the specified axis

        Parameters
        ----------
        axis_index: int
            The index along which one should sort the data
        spread_index: int
            for spread data only, specifies which spread axis to use

        Returns
        -------
        np.ndarray: the sorted index from the specified axis
        tuple of slice:
            used to slice the underlying data
        """
        ...

    @abstractmethod
    def get_axis_from_index_spread(self, index: int, spread_order: int) -> Axis:
        """in spread mode, different nav axes have the same index (but not
        the same spread_order integer value)

        """
        ...

    def compute_sig_indexes(self):
        _shape = list(self._data_shape)
        indexes = list(np.arange(len(self._data_shape)))
        for index in self.nav_indexes:
            if index in indexes:
                indexes.pop(indexes.index(index))
        return tuple(indexes)

    def _has_get_axis_from_index(self, index: int):
        """Check if the axis referred by a given data dimensionality index is present

        Returns
        -------
        bool: True if the axis has been found else False
        Axis or None: return the axis instance if has the axis else None
        """
        if index > len(self._data_shape) or index < 0:
            raise IndexError('The specified index does not correspond to any data dimension')
        for axis in self.axes:
            if axis.index == index:
                return True, axis
        return False, None

    def _manage_named_axes(self, axes, x_axis=None, y_axis=None, nav_x_axis=None, nav_y_axis=None):
        """This method make sur old style Data is still compatible, especially when using x_axis or y_axis parameters"""
        modified = False
        if x_axis is not None:
            modified = True
            index = 0
            if len(self._data_shape) == 1 and not self._has_get_axis_from_index(0)[0]:
                # in case of Data1D the x_axis corresponds to the first data dim
                index = 0
            elif len(self._data_shape) == 2 and not self._has_get_axis_from_index(1)[0]:
                # in case of Data2D the x_axis corresponds to the second data dim (columns)
                index = 1
            axes.append(Axis(x_axis.label, x_axis.units, x_axis.data, index=index))

        if y_axis is not None:

            if len(self._data_shape) == 2 and not self._has_get_axis_from_index(0)[0]:
                modified = True
                # in case of Data2D the y_axis corresponds to the first data dim (lines)
                axes.append(Axis(y_axis.label, y_axis.units, y_axis.data, index=0))

        if nav_x_axis is not None:
            if len(self.nav_indexes) > 0:
                modified = True
                # in case of DataND the y_axis corresponds to the first data dim (lines)
                axes.append(Axis(nav_x_axis.label, nav_x_axis.units, nav_x_axis.data, index=self._nav_indexes[0]))

        if nav_y_axis is not None:
            if len(self.nav_indexes) > 1:
                modified = True
                # in case of Data2D the y_axis corresponds to the first data dim (lines)
                axes.append(Axis(nav_y_axis.label, nav_y_axis.units, nav_y_axis.data, index=self._nav_indexes[1]))

        if modified:
            self._check_axis(axes)

    @property
    def shape(self) -> Tuple[int]:
        # self._data_shape = self.compute_shape_from_axes()
        return self._data_shape

    @abstractmethod
    def compute_shape_from_axes(self):
        ...

    @property
    def sig_shape(self) -> tuple:
        return tuple([self.shape[ind] for ind in self.sig_indexes])

    @property
    def nav_shape(self) -> tuple:
        return tuple([self.shape[ind] for ind in self.nav_indexes])

    def append_axis(self, axis: Axis):
        self._axes.append(axis)
        self._check_axis([axis])

    @property
    def nav_indexes(self) -> IterableType[int]:
        return self._nav_indexes

    @nav_indexes.setter
    def nav_indexes(self, nav_indexes: IterableType[int]):
        if isinstance(nav_indexes, Iterable):
            nav_indexes = tuple(nav_indexes)
            valid = True
            for index in nav_indexes:
                if index not in self.get_axes_index():
                    logger.warning('Could not set the corresponding nav_index into the data object, not enough'
                                   ' Axis declared')
                    valid = False
                    break
            if valid:
                self._nav_indexes = nav_indexes
        else:
            logger.warning('Could not set the corresponding sig_indexes into the data object, should be an iterable')
        self.sig_indexes = self.compute_sig_indexes()
        self.shape

    @property
    def sig_indexes(self) -> IterableType[int]:
        return self._sig_indexes

    @sig_indexes.setter
    def sig_indexes(self, sig_indexes: IterableType[int]):
        if isinstance(sig_indexes, Iterable):
            sig_indexes = tuple(sig_indexes)
            valid = True
            for index in sig_indexes:
                if index in self._nav_indexes:
                    logger.warning('Could not set the corresponding sig_index into the axis manager object, '
                                   'the axis is already affected to the navigation axis')
                    valid = False
                    break
                if index not in self.get_axes_index():
                    logger.warning('Could not set the corresponding nav_index into the data object, not enough'
                                   ' Axis declared')
                    valid = False
                    break
            if valid:
                self._sig_indexes = sig_indexes
        else:
            logger.warning('Could not set the corresponding sig_indexes into the data object, should be an iterable')

    @property
    def nav_axes(self) -> List[int]:
        deprecation_msg('nav_axes parameter should not be used anymore, use nav_indexes')
        return self._nav_indexes

    @nav_axes.setter
    def nav_axes(self, nav_indexes: List[int]):
        deprecation_msg('nav_axes parameter should not be used anymore, use nav_indexes')
        self.nav_indexes = nav_indexes

    def is_axis_signal(self, axis: Axis) -> bool:
        """Check if an axis is considered signal or navigation"""
        return axis.index in self._nav_indexes

    def is_axis_navigation(self, axis: Axis) -> bool:
        """Check if an axis  is considered signal or navigation"""
        return axis.index not in self._nav_indexes

    @abstractmethod
    def get_shape_from_index(self, index: int) -> int:
        """Get the data shape at the given index"""
        ...

    def get_axes_index(self) -> List[int]:
        """Get the index list from the axis objects"""
        return [axis.index for axis in self._axes]

    @abstractmethod
    def get_axis_from_index(self, index: int, create: bool = False) -> List[Axis]:
        ...

    def get_nav_axes(self) -> List[Axis]:
        """Get the navigation axes corresponding to the data

        Use get_axis_from_index for all index in self.nav_indexes, but in spread distribution, one index may
        correspond to multiple nav axes, see Spread data distribution


        """
        return list(mutils.flatten([copy.copy(self.get_axis_from_index(index, create=True))
                                    for index in self.nav_indexes]))

    def get_signal_axes(self):
        if self.sig_indexes is None:
            self._sig_indexes = tuple([int(axis.index) for axis in self.axes if axis.index not in self.nav_indexes])
        return list(mutils.flatten([copy.copy(self.get_axis_from_index(index, create=True))
                                    for index in self.sig_indexes]))

    def is_axis_signal(self, axis: Axis) -> bool:
        """Check if an axis is considered signal or navigation"""
        return axis.index in self._nav_indexes

    def is_axis_navigation(self, axis: Axis) -> bool:
        """Check if an axis  is considered signal or navigation"""
        return axis.index not in self._nav_indexes

    def __repr__(self):
        return self._get_dimension_str()

    @abstractmethod
    def _get_dimension_str(self):
        ...


class AxesManagerUniform(AxesManagerBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def compute_shape_from_axes(self):
        if len(self.axes) != 0:
            shape = []
            for ind in range(len(self.axes)):
                shape.append(len(self.get_axis_from_index(ind, create=True)[0]))
        else:
            shape = self._data_shape
        return tuple(shape)

    def get_shape_from_index(self, index: int) -> int:
        """Get the data shape at the given index"""
        if index > len(self._data_shape) or index < 0:
            raise IndexError('The specified index does not correspond to any data dimension')
        return self._data_shape[index]

    def _check_axis(self, axes: List[Axis]):
        """Check all axis to make sure of their type and make sure their data are properly referring to the data index

        See Also
        --------
        :py:meth:`Axis.create_linear_data`
        """
        for ind, axis in enumerate(axes):
            if not isinstance(axis, Axis):
                raise TypeError(f'An axis of {self.__class__.__name__} should be an Axis object')
            if self.get_shape_from_index(axis.index) != axis.size:
                warnings.warn(DataSizeWarning('The size of the axis is not coherent with the shape of the data. '
                                              'Replacing it with a linspaced version: np.array([0, 1, 2, ...])'))
                axis.size = self.get_shape_from_index(axis.index)
                axis.scaling = 1
                axis.offset = 0
                axes[ind] = axis
        self._axes = axes

    def get_axis_from_index(self, index: int, create: bool = False) -> List[Axis]:
        """Get the axis referred by a given data dimensionality index

        If the axis is absent, create a linear one to fit the data shape if parameter create is True

        Parameters
        ----------
        index: int
            The index referring to the data ndarray shape
        create: bool
            If True and the axis referred by index has not been found in axes, create one

        Returns
        -------
        List[Axis] or None: return the list of axis instance if Data has the axis (or it has been created) else None

        See Also
        --------
        :py:meth:`Axis.create_linear_data`
        """
        index = int(index)
        has_axis, axis = self._has_get_axis_from_index(index)
        if not has_axis:
            if create:
                warnings.warn(DataIndexWarning(f'The axis requested with index {index} is not present, '
                                               f'creating a linear one...'))
                axis = Axis(index=index, offset=0, scaling=1)
                axis.size = self.get_shape_from_index(index)
            else:
                warnings.warn(DataIndexWarning(f'The axis requested with index {index} is not present, returning None'))
        return [axis]

    def get_axis_from_index_spread(self, index: int, spread_order: int) -> Axis:
        """in spread mode, different nav axes have the same index (but not
        the same spread_order integer value)

        """
        return None

    def get_sorted_index(self, axis_index: int = 0, spread_index=0) -> Tuple[np.ndarray, Tuple[slice]]:
        """ Get the index to sort the specified axis

        Parameters
        ----------
        axis_index: int
            The index along which one should sort the data
        spread_index: int
            for spread data only, specifies which spread axis to use

        Returns
        -------
        np.ndarray: the sorted index from the specified axis
        tuple of slice:
            used to slice the underlying data
        """

        axes = self.get_axis_from_index(axis_index)
        if axes[0] is not None:
            sorted_index = np.argsort(axes[0].get_data())
            axes[0].data = axes[0].get_data()[sorted_index]
            slices = []
            for ind in range(len(self.shape)):
                if ind == axis_index:
                    slices.append(sorted_index)
                else:
                    slices.append(Ellipsis)
            slices = tuple(slices)
            return sorted_index, slices
        else:
            return None, None

    def _get_dimension_str(self):
        string = "("
        for nav_index in self.nav_indexes:
            string += str(self._data_shape[nav_index]) + ", "
        string = string.rstrip(", ")
        string += "|"
        for sig_index in self.sig_indexes:
            string += str(self._data_shape[sig_index]) + ", "
        string = string.rstrip(", ")
        string += ")"
        return string


class AxesManagerSpread(AxesManagerBase):
    """For this particular data category, some explanation is needed, see example below:

    Examples
    --------
    One take images data (20x30) as a function of 2 parameters, say xaxis and yaxis non-linearly spaced on a regular
    grid.

    data.shape = (150, 20, 30)
    data.nav_indexes = (0,)

    The first dimension (150) corresponds to the navigation (there are 150 non uniform data points taken)
    The  second and third could correspond to signal data, here an image of size (20x30)
    so:
    * nav_indexes is (0, )
    * sig_indexes are (1, 2)

    xaxis = Axis(name=xaxis, index=0, data...) length 150
    yaxis = Axis(name=yaxis, index=0, data...) length 150

    In fact from such a data shape the number of navigation axes in unknown . In our example, they are 2. To somehow
    keep track of some ordering in these navigation axes, one adds an attribute to the Axis object: the spread_order 
    xaxis = Axis(name=xaxis, index=0, spread_order=0, data...) length 150
    yaxis = Axis(name=yaxis, index=0, spread_order=1, data...) length 150
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_axis(self, axes: List[Axis]):
        """Check all axis to make sure of their type and make sure their data are properly referring to the data index

        """
        for axis in axes:
            if not isinstance(axis, Axis):
                raise TypeError(f'An axis of {self.__class__.__name__} should be an Axis object')
            elif len(self.nav_indexes) != 1:
                raise ValueError('Spread data should have only one specified index in self.nav_indexes')
            elif axis.index in self.nav_indexes:
                if axis.size != self._data_shape[self.nav_indexes[0]]:
                    raise DataLengthError('all navigation axes should have the same size')

    def compute_shape_from_axes(self):
        """Get data shape from axes

        First get the nav length from one of the navigation axes
        Then check for signal axes
        """
        if len(self.axes) != 0:

            axes = sorted(self.axes, key=lambda axis: axis.index)

            shape = []
            for axis in axes:
                if axis.index in self.nav_indexes:
                    shape.append(axis.size)
                    break
            for axis in axes:
                if axis.index not in self.nav_indexes:
                    shape.append(axis.size)
        else:
            shape = self._data_shape
        return tuple(shape)

    def get_shape_from_index(self, index: int) -> int:
        """Get the data shape at the given index"""
        if index > len(self._data_shape) or index < 0:
            raise IndexError('The specified index does not correspond to any data dimension')
        return self._data_shape[index]

    def get_axis_from_index(self, index: int, create: bool = False) -> List[Axis]:
        """in spread mode, different nav axes have the same index (but not
        the same spread_order integer value) so may return multiple axis

        No possible "linear" creation in this mode except if the index is a signal index

        """
        if index in self.nav_indexes:
            axes = []
            for axis in self.axes:
                if axis.index == index:
                    axes.append(axis)
            return axes
        else:
            index = int(index)
            try:
                has_axis, axis = self._has_get_axis_from_index(index)
            except IndexError:
                axis = [None]
                has_axis = False
                return axis

            if not has_axis and index in self.sig_indexes:
                if create:
                    warnings.warn(DataIndexWarning(f'The axis requested with index {index} is not present, '
                                                   f'creating a linear one...'))
                    axis = Axis(index=index, offset=0, scaling=1)
                    axis.size = self.get_shape_from_index(index)
                else:
                    warnings.warn(DataIndexWarning(f'The axis requested with index {index} is not present, returning None'))

            return [axis]

    def get_axis_from_index_spread(self, index: int, spread_order: int) -> Axis:
        """in spread mode, different nav axes have the same index (but not
        the same spread_order integer value)

        """
        for axis in self.axes:
            if axis.index == index and axis.spread_order == spread_order:
                return axis

    def get_sorted_index(self, axis_index: int = 0, spread_index=0) -> Tuple[np.ndarray, Tuple[slice]]:
        """ Get the index to sort the specified axis

        Parameters
        ----------
        axis_index: int
            The index along which one should sort the data
        spread_index: int
            for spread data only, specifies which spread axis to use

        Returns
        -------
        np.ndarray: the sorted index from the specified axis
        tuple of slice:
            used to slice the underlying data
        """

        if axis_index in self.nav_indexes:
            axis = self.get_axis_from_index_spread(axis_index, spread_index)
        else:
            axis = self.get_axis_from_index(axis_index)[0]

        if axis is not None:
            sorted_index = np.argsort(axis.get_data())
            slices = []
            for ind in range(len(self.shape)):
                if ind == axis_index:
                    slices.append(sorted_index)
                else:
                    if slices[-1] is Ellipsis:  # only one ellipsis
                        slices.append(Ellipsis)
            slices = tuple(slices)

            for nav_index in self.nav_indexes:
                for axis in self.get_axis_from_index(nav_index):
                    axis.data = axis.get_data()[sorted_index]

            return sorted_index, slices
        else:
            return None, None

    def _get_dimension_str(self):
        try:
            string = "("
            for nav_index in self.nav_indexes:
                string += str(self._data_shape[nav_index]) + ", "
                break
            string = string.rstrip(", ")
            string += "|"
            for sig_index in self.sig_indexes:
                string += str(self._data_shape[sig_index]) + ", "
            string = string.rstrip(", ")
            string += ")"
        except Exception as e:
            string = f'({self._data_shape})'
        finally:
            return string


class DataWithAxes(DataBase):
    """Data object with Axis objects corresponding to underlying data nd-arrays

    Parameters
    ----------
    axes: list of Axis
        the list of Axis object for proper plotting, calibration ...
    nav_indexes: tuple of int
        highlight which Axis in axes is Signal or Navigation axis depending on the content:
        For instance, nav_indexes = (2,), means that the axis with index 2 in a at least 3D ndarray data is the first
        navigation axis
        For instance, nav_indexes = (3,2), means that the axis with index 3 in a at least 4D ndarray data is the first
        navigation axis while the axis with index 2 is the second navigation Axis. Axes with index 0 and 1 are signal
        axes of 2D ndarray data
    """

    def __init__(self, *args, axes: List[Axis] = [], nav_indexes: Tuple[int] = (), **kwargs):

        if 'nav_axes' in kwargs:
            deprecation_msg('nav_axes parameter should not be used anymore, use nav_indexes')
            nav_indexes = kwargs.pop('nav_axes')

        x_axis = kwargs.pop('x_axis') if 'x_axis' in kwargs else None
        y_axis = kwargs.pop('y_axis') if 'y_axis' in kwargs else None

        nav_x_axis = kwargs.pop('nav_x_axis') if 'nav_x_axis' in kwargs else None
        nav_y_axis = kwargs.pop('nav_y_axis') if 'nav_y_axis' in kwargs else None

        super().__init__(*args, **kwargs)

        self._axes = axes

        other_kwargs = dict(x_axis=x_axis, y_axis=y_axis, nav_x_axis=nav_x_axis, nav_y_axis=nav_y_axis)

        self.set_axes_manager(self.shape, axes=axes, nav_indexes=nav_indexes, **other_kwargs)

        self.inav: Iterable[DataWithAxes] = SpecialSlicersData(self, True)
        self.isig: Iterable[DataWithAxes] = SpecialSlicersData(self, False)

        self.get_dim_from_data_axes()  # in DataBase, dim is processed from the shape of data, but if axes are provided
        #then use get_dim_from axes

    def set_axes_manager(self, data_shape, axes, nav_indexes, **kwargs):
        if self.distribution.name == 'uniform' or len(nav_indexes) == 0:
            self._distribution = DataDistribution['uniform']
            self.axes_manager = AxesManagerUniform(data_shape=data_shape, axes=axes, nav_indexes=nav_indexes,
                                                   **kwargs)
        elif self.distribution.name == 'spread':
            self.axes_manager = AxesManagerSpread(data_shape=data_shape, axes=axes, nav_indexes=nav_indexes,
                                                  **kwargs)
        else:
            raise ValueError(f'Such a data distribution ({data.distribution}) has no AxesManager')

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self.name} <len:{self.length}> {self._am}>'

    def sort_data(self, axis_index: int = 0, spread_index=0, inplace=False) -> DataWithAxes:
        """ Sort data along a given axis, default is 0

        Parameters
        ----------
        axis_index: int
            The index along which one should sort the data
        spread_index: int
            for spread data only, specifies which spread axis to use
        inplace: bool
            modify in place or not the data (and its axes)

        Returns
        -------
        DataWithAxes
        """
        if inplace:
            data = self
        else:
            data = self.deepcopy()
        sorted_index, slices = data._am.get_sorted_index(axis_index, spread_index)
        if sorted_index is not None:
            for ind in range(len(data)):
                data.data[ind] = data.data[ind][slices]
        return data

    def transpose(self):
        """replace the data by their transposed version

        Valid only for 2D data
        """
        if self.dim == 'Data2D':
            self.data[:] = [data.T for data in self.data]
            for axis in self.axes:
                axis.index = 0 if axis.index == 1 else 1

    def mean(self, axis: int = 0) -> DataWithAxes:
        """Process the mean of the data on the specified axis and returns the new data

        Parameters
        ----------
        axis: int

        Returns
        -------
        DataWithAxes
        """
        dat_mean = []
        for dat in self.data:
            dat_mean.append(np.mean(dat, axis=axis))
        return self.deepcopy_with_new_data(dat_mean, remove_axes_index=axis)
    
    def sum(self, axis: int = 0) -> DataWithAxes:
        """Process the sum of the data on the specified axis and returns the new data

        Parameters
        ----------
        axis: int

        Returns
        -------
        DataWithAxes
        """
        dat_sum = []
        for dat in self.data:
            dat_sum.append(np.sum(dat, axis=axis))
        return self.deepcopy_with_new_data(dat_sum, remove_axes_index=axis)
    
    def ft(self, axis: int = 0) -> DataWithAxes:
        """Process the Fourier Transform of the data on the specified axis and returns the new data

        Parameters
        ----------
        axis: int

        Returns
        -------
        DataWithAxes
        """
        dat_ft = []
        for dat in self.data:
            dat_ft.append(mutils.ft(dat, dim=axis))
        return self.deepcopy_with_new_data(dat_ft)

    def ift(self, axis: int = 0) -> DataWithAxes:
        """Process the inverse Fourier Transform of the data on the specified axis and returns the new data

        Parameters
        ----------
        axis: int

        Returns
        -------
        DataWithAxes
        """
        dat_ift = []
        for dat in self.data:
            dat_ift.append(mutils.ift(dat, dim=axis))
        return self.deepcopy_with_new_data(dat_ift)
    
    def get_dim_from_data_axes(self) -> DataDim:
        """Get the dimensionality DataDim from data taking into account nav indexes
        """
        if len(self.axes) != len(self.shape):
            self._dim = self.get_dim_from_data(self.data)
        else:
            if len(self.nav_indexes) > 0:
                self._dim = DataDim['DataND']
            else:
                if len(self.axes) == 0:
                    self._dim = DataDim['Data0D']
                elif len(self.axes) == 1:
                    self._dim = DataDim['Data1D']
                elif len(self.axes) == 2:
                    self._dim = DataDim['Data2D']
        if len(self.nav_indexes) > 0:
            self._dim = DataDim['DataND']
        return self._dim

    @property
    def n_axes(self):
        """Get the number of axes (even if not specified)"""
        return len(self.axes)

    @property
    def axes(self):
        """convenience property to fetch attribute from axis_manager"""
        return self._am.axes

    @axes.setter
    def axes(self, axes: List[Axis]):
        """convenience property to set attribute from axis_manager"""
        self.set_axes_manager(self.shape, axes=axes, nav_indexes=self.nav_indexes)

    def axes_limits(self, axes_indexes: List[int] = None) -> List[Tuple[float, float]]:
        """Get the limits of specified axes (all if axes_indexes is None)"""
        if axes_indexes is None:
            return [(axis.min(), axis.max()) for axis in self.axes]
        else:
            return [(axis.min(), axis.max()) for axis in self.axes if axis.index in axes_indexes]

    @property
    def sig_indexes(self):
        """convenience property to fetch attribute from axis_manager"""
        return self._am.sig_indexes

    @property
    def nav_indexes(self):
        """convenience property to fetch attribute from axis_manager"""
        return self._am.nav_indexes

    @nav_indexes.setter
    def nav_indexes(self, indexes: List[int]):
        """create new axis manager with new navigation indexes"""
        self.set_axes_manager(self.shape, axes=self.axes, nav_indexes=indexes)
        self.get_dim_from_data_axes()

    def get_nav_axes(self) -> List[Axis]:
        return self._am.get_nav_axes()

    def get_nav_axes_with_data(self) -> List[Axis]:
        """Get the data's navigation axes making sure there is data in the data field"""
        axes = self.get_nav_axes()
        for axis in axes:
            if axis.get_data() is None:
                axis.create_linear_data(self.shape[axis.index])
        return axes

    def get_axis_indexes(self) -> List[int]:
        """Get all present different axis indexes"""
        return sorted(list(set([axis.index for axis in self.axes])))

    def get_axis_from_index(self, index, create=False):
        return self._am.get_axis_from_index(index, create)

    def create_missing_axes(self):
        """Check if given the data shape, some axes are missing to properly define the data (especially for plotting)"""
        axes = self.axes[:]
        for index in self.nav_indexes + self.sig_indexes:
            if len(self.get_axis_from_index(index)) != 0 and self.get_axis_from_index(index)[0] is None:
                axes.extend(self.get_axis_from_index(index, create=True))
        self.axes = axes

    def _compute_slices(self, slices, is_navigation=True):
        """Compute the total slice to apply to the data

        Filling in Ellipsis when no slicing should be done
        """
        if isinstance(slices, numbers.Number) or isinstance(slices, slice):
            slices = [slices]
        if is_navigation:
            indexes = self._am.nav_indexes
        else:
            indexes = self._am.sig_indexes
        total_slices = []
        slices = list(slices)
        for ind in range(len(self.shape)):
            if ind in indexes:
                total_slices.append(slices.pop(0))
            elif len(total_slices) == 0 or total_slices[-1] != Ellipsis:
                total_slices.append(Ellipsis)
        total_slices = tuple(total_slices)
        return total_slices

    def _slicer(self, slices, is_navigation=True):
        """Apply a given slice to the data either navigation or signal dimension

        Parameters
        ----------
        slices: tuple of slice or int
            the slices to apply to the data
        is_navigation: bool
            if True apply the slices to the navigation dimension else to the signal ones

        Returns
        -------
        DataWithAxes
            Object of the same type as the initial data, derived from DataWithAxes. But with lower data size due to the
             slicing and with eventually less axes.
        """

        if isinstance(slices, numbers.Number) or isinstance(slices, slice):
            slices = [slices]
        total_slices = self._compute_slices(slices, is_navigation)
        new_arrays_data = [np.atleast_1d(np.squeeze(dat[total_slices])) for dat in self.data]
        tmp_axes = self._am.get_signal_axes() if is_navigation else self._am.get_nav_axes()
        axes_to_append = [copy.deepcopy(axis) for axis in tmp_axes]

        # axes_to_append are the axes to append to the new produced data (basically the ones to keep)

        indexes_to_get = self.nav_indexes if is_navigation else self.sig_indexes
        # indexes_to_get are the indexes of the axes where the slice should be applied

        _indexes = list(self.nav_indexes)
        _indexes.extend(self.sig_indexes)
        lower_indexes = dict(zip(_indexes, [0 for _ in range(len(_indexes))]))
        # lower_indexes will store for each *axis index* how much the index should be reduced because one axis has
        # been removed

        axes = []
        nav_indexes = [] if is_navigation else list(self._am.nav_indexes)
        for ind_slice, _slice in enumerate(slices):
            ax = self._am.get_axis_from_index(indexes_to_get[ind_slice])
            if len(ax) != 0 and ax[0] is not None:
                for ind in range(len(ax)):
                    ax[ind] = ax[ind].iaxis[_slice]

                if not(ax[0] is None or ax[0].size <= 1):  # means the slice kept part of the axis
                    if is_navigation:
                        nav_indexes.append(self._am.nav_indexes[ind_slice])
                    axes.extend(ax)
                else:
                    for axis in axes_to_append:  # means we removed one of the axes (and data dim),
                        # hence axis index above current index should be lowered by 1
                        if axis.index > indexes_to_get[ind_slice]:
                            lower_indexes[axis.index] += 1
                    for index in indexes_to_get[ind_slice+1:]:
                        lower_indexes[index] += 1

        axes.extend(axes_to_append)
        for axis in axes:
            axis.index -= lower_indexes[axis.index]
        for ind in range(len(nav_indexes)):
            nav_indexes[ind] -= lower_indexes[nav_indexes[ind]]
        data = DataWithAxes(self.name, data=new_arrays_data, nav_indexes=tuple(nav_indexes), axes=axes,
                            source='calculated', origin=self.origin,
                            labels=self.labels[:],
                            distribution=self.distribution if len(nav_indexes) != 0 else DataDistribution['uniform'])
        return data

    def deepcopy_with_new_data(self, data: List[np.ndarray] = None,
                               remove_axes_index: List[int] = None,
                               source: DataSource = 'calculated',
                               keep_dim=False) -> DataWithAxes:
        """deepcopy without copying the initial data (saving memory)

        The new data, may have some axes stripped as specified in remove_axes_index

        Parameters
        ----------
        data: list of numpy ndarray
            The new data
        remove_axes_index: tuple of int
            indexes of the axis to be removed
        source: DataSource
        keep_dim: bool
            if False (the default) will calculate the new dim based on the data shape
            else keep the same (be aware it could lead to issues)

        Returns
        -------
        DataWithAxes
        """
        try:
            old_data = self.data
            self._data = None
            new_data = self.deepcopy()
            new_data._data = data
            new_data.get_dim_from_data(data)

            if source is not None:
                source = enum_checker(DataSource, source)
                new_data._source = source


            if remove_axes_index is not None:
                if not isinstance(remove_axes_index, Iterable):
                    remove_axes_index = [remove_axes_index]
                    
                lower_indexes = dict(zip(new_data.get_axis_indexes(),
                                         [0 for _ in range(len(new_data.get_axis_indexes()))]))
                # lower_indexes will store for each *axis index* how much the index should be reduced because one axis has
                # been removed

                nav_indexes = list(new_data.nav_indexes)
                sig_indexes = list(new_data.sig_indexes)
                for index in remove_axes_index:
                    for axis in new_data.get_axis_from_index(index):
                        new_data.axes.remove(axis)

                    if index in new_data.nav_indexes:
                        nav_indexes.pop(nav_indexes.index(index))
                    if index in new_data.sig_indexes:
                        sig_indexes.pop(sig_indexes.index(index))

                    # for ind, nav_ind in enumerate(nav_indexes):
                    #     if nav_ind > index and nav_ind not in remove_axes_index:
                    #         nav_indexes[ind] -= 1

                    # for ind, sig_ind in enumerate(sig_indexes):
                    #     if sig_ind > index:
                    #         sig_indexes[ind] -= 1
                    for axis in new_data.axes:
                        if axis.index > index and axis.index not in remove_axes_index:
                            lower_indexes[axis.index] += 1

                for axis in new_data.axes:
                    axis.index -= lower_indexes[axis.index]
                for ind in range(len(nav_indexes)):
                    nav_indexes[ind] -= lower_indexes[nav_indexes[ind]]

                new_data.nav_indexes = tuple(nav_indexes)
                # new_data._am.sig_indexes = tuple(sig_indexes)

            new_data._shape = data[0].shape
            if not keep_dim:
                new_data._dim = self._get_dim_from_data(data)
            return new_data

        except Exception as e:
            pass
        finally:
            self._data = old_data

    def deepcopy(self):
        return copy.deepcopy(self)

    @property
    def _am(self) -> AxesManagerBase:
        return self.axes_manager

    def get_data_dimension(self) -> str:
        return str(self._am)


class DataRaw(DataWithAxes):
    """Specialized DataWithAxes set with source as 'raw'. To be used for raw data"""
    def __init__(self, *args,  **kwargs):
        if 'source' in kwargs:
            kwargs.pop('source')
        super().__init__(*args, source=DataSource['raw'], **kwargs)


class DataActuator(DataRaw):
    """Specialized DataWithAxes set with source as 'raw'. To be used for raw data generated by actuator plugins"""
    def __init__(self, *args, **kwargs):
        if len(args) == 0 and 'name' not in kwargs:
            args = ['actuator']
        if 'data' not in kwargs:
            kwargs['data'] = [np.array([0.])]
        elif isinstance(kwargs['data'], numbers.Number):  # useful formatting
            kwargs['data'] = [np.array([kwargs['data']])]
        super().__init__(*args, **kwargs)

    def __repr__(self):
        if self.dim.name == 'Data0D':
            return f'{self.__class__.__name__} <{self.data[0][0]}>'
        else:
            return f'{self.__class__.__name__} <{self.shape}>>'

    def value(self):
        """Returns the underlying float value if this data holds only a float otherwise returns a mean of the
        underlying data"""
        if self.length == 1 and self.size == 1:
            return float(self.data[0][0])
        else:
            return float(np.mean(self.data))


class DataFromPlugins(DataRaw):
    """Specialized DataWithAxes set with source as 'raw'. To be used for raw data generated by Detector plugins

    It introduces by default to extra attributes, plot and save. Their presence can be checked in the
    extra_attributes list.

    Parameters
    ----------
    plot: bool
        If True the underlying data will be plotted in the DAQViewer
    save: bool
        If True the underlying data will be saved

    Attributes
    ----------
    plot: bool
        If True the underlying data will be plotted in the DAQViewer
    save: bool
        If True the underlying data will be saved
    """
    def __init__(self, *args, **kwargs):
        if 'plot' not in kwargs:
            kwargs['plot'] = True
        if 'save' not in kwargs:
            kwargs['save'] = True
        super().__init__(*args, **kwargs)


class DataCalculated(DataWithAxes):
    """Specialized DataWithAxes set with source as 'calculated'. To be used for processed/calculated data"""
    def __init__(self, *args, axes=[],  **kwargs):
        if 'source' in kwargs:
            kwargs.pop('source')
        super().__init__(*args, source=DataSource['calculated'], axes=axes, **kwargs)


class DataFromRoi(DataCalculated):
    """Specialized DataWithAxes set with source as 'calculated'.To be used for processed data from region of interest"""
    def __init__(self, *args, axes=[], **kwargs):
        super().__init__(*args, axes=axes, **kwargs)


class DataToExport(DataLowLevel):
    """Object to store all raw and calculated DataWithAxes data for later exporting, saving, sending signal...

    Includes methods to retrieve data from dim, source...
    Stored data have a unique identifier their name. If some data is appended with an existing name, it will replace
    the existing data. So if you want to append data that has the same name

    Parameters
    ----------
    name: str
        The identifier of the exporting object
    data: list of DataWithAxes
        All the raw and calculated data to be exported

    Attributes
    ----------
    name
    timestamp
    data
    """

    def __init__(self, name: str, data: List[DataWithAxes] = [], **kwargs):
        """

        Parameters
        ----------
        name
        data
        """
        super().__init__(name)
        if not isinstance(data, list):
            raise TypeError('Data stored in a DataToExport object should be as a list of objects'
                            ' inherited from DataWithAxis')
        self._data = []

        self.data = data
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def affect_name_to_origin_if_none(self):
        """Affect self.name to all DataWithAxes children's attribute origin if this origin is not defined"""
        for dat in self.data:
            if dat.origin is None or dat.origin == '':
                dat.origin = self.name

    def __sub__(self, other: object):
        if isinstance(other, DataToExport) and len(other) == len(self):
            new_data = copy.deepcopy(self)
            for ind_dfp in range(len(self)):
                new_data[ind_dfp] = self[ind_dfp] - other[ind_dfp]
            return new_data
        else:
            raise TypeError(f'Could not substract a {other.__class__.__name__} or a {self.__class__.__name__} '
                            f'of a different length')

    def __add__(self, other: object):
        if isinstance(other, DataToExport) and len(other) == len(self):
            new_data = copy.deepcopy(self)
            for ind_dfp in range(len(self)):
                new_data[ind_dfp] = self[ind_dfp] + other[ind_dfp]
            return new_data
        else:
            raise TypeError(f'Could not add a {other.__class__.__name__} or a {self.__class__.__name__} '
                            f'of a different length')

    def __mul__(self, other: object):
        if isinstance(other, numbers.Number):
            new_data = copy.deepcopy(self)
            for ind_dfp in range(len(self)):
                new_data[ind_dfp] = self[ind_dfp] * other
            return new_data
        else:
            raise TypeError(f'Could not multiply a {other.__class__.__name__} with a {self.__class__.__name__} '
                            f'of a different length')

    def __truediv__(self, other: object):
        if isinstance(other, numbers.Number):
            return self * (1 / other)
        else:
            raise TypeError(f'Could not divide a {other.__class__.__name__} with a {self.__class__.__name__} '
                            f'of a different length')

    def average(self, other: DataToExport, weight: int) -> DataToExport:
        """ Compute the weighted average between self and other DataToExport and attributes it to self

        Parameters
        ----------
        other: DataToExport
        weight: int
            The weight the 'other_data' holds with respect to self

        """
        if isinstance(other, DataToExport) and len(other) == len(self):
            new_data = copy.copy(self)
            for ind_dfp in range(len(self)):
                new_data[ind_dfp] = self[ind_dfp].average(other[ind_dfp], weight)
            return new_data
        else:
            raise TypeError(f'Could not average a {other.__class__.__name__} with a {self.__class__.__name__} '
                            f'of a different length')

    def merge_as_dwa(self, dim: DataDim, name: str = None) -> DataRaw:
        """ attempt to merge all dwa into one

        Only possible if all dwa and underlying data have same shape
        """
        dim = enum_checker(DataDim, dim)
        if name is None:
            name = self.name
        filtered_data = self.get_data_from_dim(dim)
        ndarrays = []
        for dwa in filtered_data:
            ndarrays.extend(dwa.data)
        dwa = DataRaw(name, dim=dim, data=ndarrays)
        return dwa

    def __repr__(self):
        repr = f'{self.__class__.__name__}: {self.name} <len:{len(self)}>\n'
        for dwa in self:
            repr += f'    * {str(dwa)}\n'
        return repr

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self) -> DataWithAxes:
        if self._iter_index < len(self):
            self._iter_index += 1
            return self.data[self._iter_index-1]
        else:
            raise StopIteration

    def __getitem__(self, item) -> Union[DataWithAxes, DataToExport]:
        if isinstance(item, int) and 0 <= item < len(self):
            return self.data[item]
        elif isinstance(item, slice):
            return DataToExport(self.name, data=[self[ind] for ind in list(range(len(self))[item])])
        else:
            raise IndexError(f'The index should be a positive integer lower than the data length')

    def __setitem__(self, key, value: DataWithAxes):
        if isinstance(key, int) and 0 <= key < len(self) and isinstance(value, DataWithAxes):
            self.data[key] = value
        else:
            raise IndexError(f'The index should be a positive integer lower than the data length')

    def get_names(self, dim: DataDim = None) -> List[str]:
        """Get the names of the stored DataWithAxes,  eventually filtered by dim

        Parameters
        ----------
        dim: DataDim or str

        Returns
        -------
        list of str: the names of the (filtered) DataWithAxes data
        """
        if dim is None:
            return [data.name for data in self.data]
        else:
            return [data.name for data in self.get_data_from_dim(dim).data]

    def get_full_names(self, dim: DataDim = None):
        """Get the ful names including the origin attribute into the returned value,  eventually filtered by dim

        Parameters
        ----------
        dim: DataDim or str

        Returns
        -------
        list of str: the names of the (filtered) DataWithAxes data constructed as : origin/name

        Examples
        --------
        d0 = DataWithAxes(name='datafromdet0', origin='det0')
        """
        if dim is None:
            return [data.get_full_name() for data in self.data]
        else:
            return [data.get_full_name() for data in self.get_data_from_dim(dim).data]

    def get_data_from_full_name(self, full_name: str, deepcopy=False) -> DataWithAxes:
        """Get the DataWithAxes with matching full name"""
        if deepcopy:
            data = self.get_data_from_name_origin(full_name.split('/')[1], full_name.split('/')[0]).deepcopy()
        else:
            data = self.get_data_from_name_origin(full_name.split('/')[1], full_name.split('/')[0])
        return data

    def get_data_from_full_names(self, full_names: List[str], deepcopy=False) -> DataToExport:
        data = [self.get_data_from_full_name(full_name, deepcopy) for full_name in full_names]
        return DataToExport(name=self.name, data=data)

    def get_dim_presents(self) -> List[str]:
        dims = []
        for dim in DataDim.names():
            if len(self.get_data_from_dim(dim)) != 0:
                dims.append(dim)

        return dims

    def get_data_from_source(self, source: DataSource, deepcopy=False) -> DataToExport:
        """Get the data matching the given DataSource

        Returns
        -------
        DataToExport: filtered with data matching the dimensionality
        """
        source = enum_checker(DataSource, source)
        return self.get_data_from_attribute('source', source, deepcopy=deepcopy)

    def get_data_from_missing_attribute(self, attribute: str, deepcopy=False) -> DataToExport:
        """ Get the data matching a given attribute value

        Parameters
        ----------
        attribute: str
            a string of a possible attribute
        deepcopy: bool
            if True the returned DataToExport will contain deepcopies of the DataWithAxes
        Returns
        -------
        DataToExport: filtered with data missing the given attribute
        """
        if deepcopy:
            return DataToExport(self.name, data=[dwa.deepcopy() for dwa in self if not hasattr(dwa, attribute)])
        else:
            return DataToExport(self.name, data=[dwa for dwa in self if not hasattr(dwa, attribute)])

    def get_data_from_attribute(self, attribute: str, attribute_value: Any, deepcopy=False) -> DataToExport:
        """Get the data matching a given attribute value

        Returns
        -------
        DataToExport: filtered with data matching the attribute presence and value
        """
        selection = find_objects_in_list_from_attr_name_val(self.data, attribute, attribute_value,
                                                            return_first=False)
        selection.sort(key=lambda elt: elt[0].name)
        if deepcopy:
            data = [sel[0].deepcopy() for sel in selection]
        else:
            data = [sel[0] for sel in selection]
        return DataToExport(name=self.name, data=data)

    def get_data_from_dim(self, dim: DataDim, deepcopy=False) -> DataToExport:
        """Get the data matching the given DataDim

        Returns
        -------
        DataToExport: filtered with data matching the dimensionality
        """
        dim = enum_checker(DataDim, dim)
        return self.get_data_from_attribute('dim', dim, deepcopy=deepcopy)

    def get_data_from_dims(self, dims: List[DataDim], deepcopy=False) -> DataToExport:
        """Get the data matching the given DataDim

        Returns
        -------
        DataToExport: filtered with data matching the dimensionality
        """
        data = DataToExport(name=self.name)
        for dim in dims:
            data.append(self.get_data_from_dim(dim, deepcopy=deepcopy))
        return data

    def get_data_from_sig_axes(self, Naxes: int, deepcopy: bool = False) -> DataToExport:
        """Get the data matching the given number of signal axes

        Parameters
        ----------
        Naxes: int
            Number of signal axes in the DataWithAxes objects

        Returns
        -------
        DataToExport: filtered with data matching the number of signal axes
        """
        data = DataToExport(name=self.name)
        for _data in self:
            if len(_data.sig_indexes) == Naxes:
                if deepcopy:
                    data.append(_data.deepcopy())
                else:
                    data.append(_data)
        return data

    def get_data_from_Naxes(self, Naxes: int, deepcopy: bool = False) -> DataToExport:
        """Get the data matching the given number of axes

        Parameters
        ----------
        Naxes: int
            Number of axes in the DataWithAxes objects

        Returns
        -------
        DataToExport: filtered with data matching the number of axes
        """
        data = DataToExport(name=self.name)
        for _data in self:
            if len(_data.shape) == Naxes:
                if deepcopy:
                    data.append(_data.deepcopy())
                else:
                    data.append(_data)
        return data

    def get_data_with_naxes_lower_than(self, n_axes=2, deepcopy: bool = False) -> DataToExport:
        """Get the data with n axes lower than the given number

        Parameters
        ----------
        Naxes: int
            Number of axes in the DataWithAxes objects

        Returns
        -------
        DataToExport: filtered with data matching the number of axes
        """
        data = DataToExport(name=self.name)
        for _data in self:
            if _data.n_axes <= n_axes:
                if deepcopy:
                    data.append(_data.deepcopy())
                else:
                    data.append(_data)
        return data

    def get_data_from_name(self, name: str) -> DataWithAxes:
        """Get the data matching the given name"""
        data, _ = find_objects_in_list_from_attr_name_val(self.data, 'name', name, return_first=True)
        return data

    def get_data_from_names(self, names: List[str]) -> DataToExport:
        return DataToExport(self.name, data=[dwa for dwa in self if dwa.name in names])

    def get_data_from_name_origin(self, name: str, origin: str = '') -> DataWithAxes:
        """Get the data matching the given name and the given origin"""
        if origin == '':
            data, _ = find_objects_in_list_from_attr_name_val(self.data, 'name', name, return_first=True)
        else:
            selection = find_objects_in_list_from_attr_name_val(self.data, 'name', name, return_first=False)
            selection = [sel[0] for sel in selection]
            data, _ = find_objects_in_list_from_attr_name_val(selection, 'origin', origin)
        return data

    def index(self, data: DataWithAxes):
        return self.data.index(data)

    def index_from_name_origin(self, name: str, origin: str = '') -> List[DataWithAxes]:
        """Get the index of a given DataWithAxes within the list of data"""
        """Get the data matching the given name and the given origin"""
        if origin == '':
            _, index = find_objects_in_list_from_attr_name_val(self.data, 'name', name, return_first=True)
        else:
            selection = find_objects_in_list_from_attr_name_val(self.data, 'name', name, return_first=False)
            data_selection = [sel[0] for sel in selection]
            index_selection = [sel[1] for sel in selection]
            _, index = find_objects_in_list_from_attr_name_val(data_selection, 'origin', origin)
            index = index_selection[index]
        return index

    def pop(self, index: int) -> DataWithAxes:
        """return and remove the DataWithAxes referred by its index

        Parameters
        ----------
        index: int
            index as returned by self.index_from_name_origin

        See Also
        --------
        index_from_name_origin
        """
        return self.data.pop(index)

    def remove(self, dwa: DataWithAxes):
        return self.pop(self.data.index(dwa))

    @property
    def data(self) -> List[DataWithAxes]:
        """List[DataWithAxes]: get the data contained in the object"""
        return self._data

    @data.setter
    def data(self, new_data: List[DataWithAxes]):
        for dat in new_data:
            self._check_data_type(dat)
        self._data[:] = [dat for dat in new_data]  # shallow copyto make sure that if the original list
        # is changed, the change will not be applied in here

        self.affect_name_to_origin_if_none()

    @staticmethod
    def _check_data_type(data: DataWithAxes):
        """Make sure data is a DataWithAxes object or inherited"""
        if not isinstance(data, DataWithAxes):
            raise TypeError('Data stored in a DataToExport object should be objects inherited from DataWithAxis')

    def deepcopy(self):
        return DataToExport('Copy', data=[data.deepcopy() for data in self])

    @dispatch(list)
    def append(self, data_list: List[DataWithAxes]):
        for dwa in data_list:
            self.append(dwa)

    @dispatch(DataWithAxes)
    def append(self, dwa: DataWithAxes):
        """Append/replace DataWithAxes object to the data attribute

        Make sure only one DataWithAxes object with a given name is in the list except if they don't have the same
        origin identifier
        """
        dwa = dwa.deepcopy()
        self._check_data_type(dwa)
        obj = self.get_data_from_name_origin(dwa.name, dwa.origin)
        if obj is not None:
            self._data.pop(self.data.index(obj))
        self._data.append(dwa)

    @dispatch(object)
    def append(self, dte: DataToExport):
        if isinstance(dte, DataToExport):
            self.append(dte.data)


class DataScan(DataToExport):
    """Specialized DataToExport.To be used for data to be saved """
    def __init__(self, name: str, data: List[DataWithAxes] = [], **kwargs):
        super().__init__(name, data, **kwargs)


if __name__ == '__main__':


    d1 = DataFromRoi(name=f'Hlineout_', data=[np.zeros((24,))],
                     x_axis=Axis(data=np.zeros((24,)), units='myunits', label='mylabel1'))
    d2 = DataFromRoi(name=f'Hlineout_', data=[np.zeros((12,))],
                     x_axis=Axis(data=np.zeros((12,)),
                                 units='myunits2',
                                 label='mylabel2'))

    Nsig = 200
    Nnav = 10
    x = np.linspace(-Nsig/2, Nsig/2-1, Nsig)

    dat = np.zeros((Nnav, Nsig))
    for ind in range(Nnav):
        dat[ind] = mutils.gauss1D(x,  50 * (ind -Nnav / 2), 25 / np.sqrt(2))

    data = DataRaw('mydata', data=[dat], nav_indexes=(0,),
                   axes=[Axis('nav', data=np.linspace(0, Nnav-1, Nnav), index=0),
                         Axis('sig', data=x, index=1)])

    data2 = copy.copy(data)

    data3 = data.deepcopy_with_new_data([np.sum(dat, 1)], remove_axes_index=(1,))

    print('done')




# -*- coding: utf-8 -*-
"""
Created on Fri Aug 30 12:21:56 2019

@author: Weber
"""
from collections import OrderedDict
import select
from typing import List
import socket

import numpy as np
from qtpy.QtCore import QObject, Signal, Slot, QThread
from qtpy import QtWidgets

from pymodaq.utils.parameter import utils as putils
from pymodaq.utils.parameter import ioxml
from pymodaq.utils.daq_utils import getLineInfo, ThreadCommand
from pymodaq.utils.data import DataFromPlugins, DataActuator
from pymodaq.utils import math_utils as mutils
from pymodaq.utils.config import Config
from pymodaq.utils.parameter import Parameter
from pymodaq.utils.data import DataToExport
from pymodaq.utils.tcp_ip.mysocket import Socket
from pymodaq.utils.tcp_ip.serializer import Serializer, DeSerializer

config = Config()

tcp_parameters = [
    {'title': 'Port:', 'name': 'port_id', 'type': 'int', 'value': config('network', 'tcp-server', 'port'), },
    {'title': 'IP:', 'name': 'socket_ip', 'type': 'str', 'value': config('network', 'tcp-server', 'ip'), },
    {'title': 'Settings PyMoDAQ Client:', 'name': 'settings_client', 'type': 'group', 'children': []},
    {'title': 'Infos Client:', 'name': 'infos', 'type': 'group', 'children': []},
    {'title': 'Connected clients:', 'name': 'conn_clients', 'type': 'table',
     'value': dict(), 'header': ['Type', 'adress']}, ]


class TCPClientTemplate:
    params = []

    def __init__(self, ipaddress="192.168.1.62", port=6341, params_state=None, client_type=""):
        """Create a socket client

        Parameters
        ----------
        ipaddress: (str) the IP address of the server
        port: (int) the port where to communicate with the server
        params_state: (dict) state of the Parameter settings of the module instantiating this client and wishing to
                            export its settings to the server. Obtained from param.saveState() where param is an
                            instance of Parameter object, see pyqtgraph.parametertree::Parameter
        client_type: (str) should be one of the accepted client_type by the TCPServer instance (within pymodaq it is
                            either 'GRABBER' or 'ACTUATOR'
        """
        super().__init__()

        self.ipaddress = ipaddress
        self.port = port
        self._socket: Socket = None
        self._deserializer: DeSerializer = None
        self.connected = False
        self.settings = Parameter.create(name='Settings', type='group', children=self.params)
        if params_state is not None:
            if isinstance(params_state, dict):
                self.settings.restoreState(params_state)
            elif isinstance(params_state, Parameter):
                self.settings.restoreState(params_state.saveState())

        self.client_type = client_type

    @property
    def socket(self) -> Socket:
        return self._socket

    @socket.setter
    def socket(self, sock: Socket):
        self._socket = sock
        self._deserializer = DeSerializer(sock)

    def close(self):
        if self.socket is not None:
            self.socket.close()

    def _connect_socket(self):
        # create an INET, STREAMing socket
        self.socket = Socket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
        # now connect to the web server on port 80 - the normal http port
        self.socket.connect((self.ipaddress, self.port))

    def init_connection(self, extra_commands=[]):
        """init the socket connection then call the post_init method where to place custom initialization"""
        try:
            self._connect_socket()
            self.post_init(extra_commands)
            self.connected = True
            self.poll_connection()

        except ConnectionRefusedError as e:
            self.not_connected(e)
            self.connected = False

    def poll_connection(self):
        while True:
            try:
                ready_to_read, ready_to_write, in_error = \
                    select.select([self.socket.socket], [self.socket.socket], [self.socket.socket], 0)

                if len(ready_to_read) != 0:
                    self.ready_to_read()

                if len(in_error) != 0:
                    self.ready_with_error()

                if  len(ready_to_write) != 0:
                    self.ready_to_write()

                QtWidgets.QApplication.processEvents()

            except Exception as e:
                self.process_error_in_polling(e)
                break

    def not_connected(self, e: ConnectionRefusedError):
        raise NotImplementedError

    def ready_to_read(self):
        """Do stuff (like read data) when messages arrive through the socket"""
        raise NotImplementedError

    def ready_to_write(self):
        """Send stuff into the socket"""
        raise NotImplementedError

    def ready_with_error(self):
        """Error in the socket communication"""
        raise NotImplementedError

    def process_error_in_polling(self, e: Exception):
        raise NotImplementedError

    def post_init(self, extra_commands=[]):
        """To implement in a real object implementation"""
        raise NotImplementedError


class TCPClient(TCPClientTemplate, QObject):
    """
    PyQt5 object initializing a TCP socket client. Can be used by any module but is a builtin functionality of all
    actuators and detectors of PyMoDAQ

    The module should init TCPClient, move it in a thread and communicate with it using a custom signal connected to
    TCPClient.queue_command slot. The module should also connect TCPClient.cmd_signal to one of its methods inorder to
    get info/data back from the client

    The client itself communicate with a TCP server, it is best to use a server object subclassing the TCPServer
    class defined within this python module

    """
    cmd_signal = Signal(ThreadCommand)  # signal to connect with a module slot in order to start communication back
    params = []

    def __init__(self, ipaddress="192.168.1.62", port=6341, params_state=None, client_type="GRABBER"):
        """Create a socket client particularly fit to be used with PyMoDAQ's TCPServer

        Parameters
        ----------
        ipaddress: (str) the IP address of the server
        port: (int) the port where to communicate with the server
        params_state: (dict) state of the Parameter settings of the module instantiating this client and wishing to
                            export its settings to the server. Obtained from param.saveState() where param is an
                            instance of Parameter object, see pyqtgraph.parametertree::Parameter
        client_type: (str) should be one of the accepted client_type by the TCPServer instance (within pymodaq it is
                            either 'GRABBER' or 'ACTUATOR'
        """
        QObject.__init__(self)
        TCPClientTemplate.__init__(self, ipaddress, port, params_state, client_type)

    def send_data(self, data: DataToExport):
        # first send 'Done' and then send the length of the list
        if not isinstance(data, DataToExport):
            raise TypeError(f'should send a DataToExport object')
        if self.socket is not None:
            self.socket.check_sended_with_serializer('Done')
            self.socket.check_sended_with_serializer(data)

    def send_infos_xml(self, infos: str):
        if self.socket is not None:
            self.socket.check_sended_with_serializer('Infos')
            self.socket.check_sended_with_serializer(infos)

    def send_info_string(self, info_to_display, value_as_string):
        if self.socket is not None:
            self.socket.check_sended_with_serializer('Info')
            self.socket.check_sended_with_serializer(info_to_display)  # the actual info to display as a string
            if not isinstance(value_as_string, str):
                value_as_string = str(value_as_string)
            self.socket.check_sended_with_serializer(value_as_string)

    @Slot(ThreadCommand)
    def queue_command(self, command=ThreadCommand):
        """
        when this TCPClient object is within a thread, the corresponding module communicate with it with signal and slots
        from module to client: module_signal to queue_command slot
        from client to module: self.cmd_signal to a module slot
        """
        if command.command == "ini_connection":
            status = self.init_connection()

        elif command.command == "quit":
            try:
                self.socket.close()
            except Exception as e:
                pass
            finally:
                self.cmd_signal.emit(ThreadCommand('disconnected'))

        elif command.command == 'update_connection':
            self.ipaddress = command.attribute['ipaddress']
            self.port = command.attribute['port']

        elif command.command == 'data_ready':
            self.data_ready(command.attribute)

        elif command.command == 'send_info':
            if self.socket is not None:
                path = command.attribute['path']
                param = command.attribute['param']

                self.socket.check_sended_with_serializer('Info_xml')
                self.socket.check_sended_with_serializer(path)

                # send value
                data = ioxml.parameter_to_xml_string(param)
                self.socket.check_sended_with_serializer(data)

        elif command.command == 'position_is':
            if self.socket is not None:
                self.socket.check_sended_with_serializer('position_is')
                self.socket.check_sended_with_serializer(command.attribute[0])

        elif command.command == 'move_done':
            if self.socket is not None:
                self.socket.check_sended_with_serializer('move_done')
                self.socket.check_sended_with_serializer(command.attribute[0])

        elif command.command == 'x_axis':
            raise DeprecationWarning('Getting axis though TCPIP is deprecated use the data objects directly')

        elif command.command == 'y_axis':
            raise DeprecationWarning('Getting axis though TCPIP is deprecated use the data objects directly')
        else:
            raise IOError('Unknown TCP client command')

    def not_connected(self, e):
        self.connected = False
        self.cmd_signal.emit(ThreadCommand('disconnected'))
        self.cmd_signal.emit(ThreadCommand('Update_Status', [getLineInfo() + str(e), 'log']))

    def ready_to_read(self):
        message = self._deserializer.string_deserialization()
        self.get_data(message)

    def ready_to_write(self):
        pass

    def ready_with_error(self):
        self.connected = False
        self.cmd_signal.emit(ThreadCommand('disconnected'))

    def process_error_in_polling(self, e: Exception):
        try:
            self.cmd_signal.emit(ThreadCommand('Update_Status', [getLineInfo() + str(e), 'log']))
            self.socket.check_sended_with_serializer('Quit')
            self.socket.close()
        except Exception:  # pragma: no cover
            pass

    def post_init(self, extra_commands=[]):

        self.cmd_signal.emit(ThreadCommand('connected'))
        self.socket.check_sended_with_serializer(self.client_type)

        self.send_infos_xml(ioxml.parameter_to_xml_string(self.settings))
        for command in extra_commands:
            if isinstance(command, ThreadCommand):
                self.cmd_signal.emit(command)

    def get_data(self, message: str):
        """

        Parameters
        ----------
        message

        Returns
        -------

        """
        if self.socket is not None:
            messg = ThreadCommand(message)

            if message == 'set_info':
                path = self._deserializer.list_deserialization()
                param_xml = self._deserializer.string_deserialization()
                messg.attribute = [path, param_xml]

            elif message == 'move_abs' or message == 'move_rel':
                position = self._deserializer.dwa_deserialization()
                messg.attribute = [position]

            self.cmd_signal.emit(messg)

    def data_ready(self, data: DataToExport):
        self.send_data(data)


class TCPServer(QObject):
    """
    Abstract class to be used as inherited by DAQ_Viewer_TCP or DAQ_Move_TCP
    """

    def __init__(self, client_type='GRABBER'):
        QObject.__init__(self)
        self.serversocket: Socket = None
        self.connected_clients = []
        self.listening = True
        self.processing = False
        self.client_type = client_type

    def close_server(self):
        """
            close the current opened server.
            Update the settings tree consequently.

            See Also
            --------
            set_connected_clients_table, daq_utils.ThreadCommand
        """
        server_socket = self.find_socket_within_connected_clients('server')
        self.remove_client(server_socket)

    def init_server(self):
        self.emit_status(ThreadCommand("Update_Status", [
            "Started new server for {:s}:{:d}".format(self.settings['socket_ip'],
                                                      self.settings['port_id']), 'log']))
        serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.serversocket = Socket(serversocket)
        # bind the socket to a public host, and a well-known port
        try:
            self.serversocket.bind(
                (self.settings['socket_ip'], self.settings['port_id']))
            # self.serversocket.bind((socket.gethostname(), self.settings.child(('port_id')).value()))
        except socket.error as msg:
            self.emit_status(ThreadCommand("Update_Status",
                                           ['Bind failed. Error Code : ' + str(msg.errno) + ' Message ' + msg.strerror,
                                            'log']))
            raise ConnectionError('Bind failed. Error Code : ' + str(msg.errno) + ' Message ' + msg.strerror)

        self.serversocket.listen(1)
        self.connected_clients.append(dict(socket=self.serversocket, type='server'))
        self.settings.child('conn_clients').setValue(self.set_connected_clients_table())

        self.timer = self.startTimer(100)  # Timer event fired every 100ms
        # self.listen_client()

    def timerEvent(self, event):
        """
            Called by set timers.
            If the process is free, start the listen_client function.

            =============== ==================== ==============================================
            **Parameters**   **Type**              **Description**

            *event*          QTimerEvent object    Containing id from timer issuing this event
            =============== ==================== ==============================================

            See Also
            --------
            listen_client
        """
        if not self.processing:
            self.listen_client()

    def find_socket_within_connected_clients(self, client_type) -> Socket:
        """
            Find a socket from a connected client with socket type corresponding.

            =============== =========== ================================
            **Parameters**    **Type**    **Description**
            *client_type*      string     The corresponding client type
            =============== =========== ================================

            Returns
            -------
            dictionnary
                the socket dictionnary
        """
        res = None
        for socket_dict in self.connected_clients:
            if socket_dict['type'] == client_type:
                res = socket_dict['socket']
        return res

    def find_socket_type_within_connected_clients(self, sock):
        """
            Find a socket type from a connected client with socket content corresponding.

            =============== =========== ===================================
            **Parameters**    **Type**   **Description**
            *sock*             ???       The socket content corresponding.
            =============== =========== ===================================

            Returns
            -------
            dictionnary
                the socket dictionnary
        """
        res = None
        for socket_dict in self.connected_clients:
            if socket_dict['socket'] == sock:
                res = socket_dict['type']
        return res

    def set_connected_clients_table(self):
        """

        """
        con_clients = OrderedDict()
        for socket_dict in self.connected_clients:
            try:
                address = str(socket_dict['socket'].getsockname())
            except Exception:
                address = "unconnected invalid socket"
            con_clients[socket_dict['type']] = address
        return con_clients

    @Slot(list)
    def print_status(self, status):
        """
            Print the given status.

            =============== ============= ================================================
            **Parameters**    **Type**       **Description**
            *status*          string list    a string list representing the status socket
            =============== ============= ================================================
        """
        print(status)

    def remove_client(self, sock):
        sock_type = self.find_socket_type_within_connected_clients(sock)
        if sock_type is not None:
            self.connected_clients.remove(dict(socket=sock, type=sock_type))
            self.settings.child('conn_clients').setValue(self.set_connected_clients_table())
            try:
                sock.close()
            except Exception:
                pass
            self.emit_status(ThreadCommand("Update_Status", ['Client ' + sock_type + ' disconnected', 'log']))

    def select(self, rlist, wlist=[], xlist=[], timeout=0):
        """
        Implements the select method, https://docs.python.org/3/library/select.html
        Parameters
        ----------
        rlist: (list) wait until ready for reading
        wlist: (list) wait until ready for writing
        xlist: (list)  wait for an “exceptional condition”
        timeout: (float) optional timeout argument specifies a time-out as a floating point number in seconds.
                When the timeout argument is omitted the function blocks until at least one file descriptor is ready.
                A time-out value of zero specifies a poll and never blocks.

        Returns
        -------
        list: readable sockets
        list: writable sockets
        list: sockets with error pending
        """

        read_sockets, write_sockets, error_sockets = select.select([sock.socket for sock in rlist],
                                                                   [sock.socket for sock in wlist],
                                                                   [sock.socket for sock in xlist],
                                                                   timeout)

        return ([Socket(sock) for sock in read_sockets], [Socket(sock) for sock in write_sockets],
                [Socket(sock) for sock in error_sockets])

    def listen_client(self):
        """
            Server function.
            Used to connect or listen incoming message from a client.
        """
        try:
            self.processing = True
            # QtWidgets.QApplication.processEvents() #to let external commands in
            read_sockets, write_sockets, error_sockets = self.select(
                [client['socket'] for client in self.connected_clients], [],
                [client['socket'] for client in self.connected_clients],
                0)
            for sock in error_sockets:
                self.remove_client(sock)

            for sock in read_sockets:
                QThread.msleep(100)
                if sock == self.serversocket:  # New connection
                    # means a new socket (client) try to reach the server
                    (client_socket, address) = self.serversocket.accept()
                    DAQ_type = DeSerializer(client_socket).string_deserialization()
                    if DAQ_type not in self.socket_types:
                        self.emit_status(ThreadCommand("Update_Status", [DAQ_type + ' is not a valid type', 'log']))
                        client_socket.close()
                        break

                    self.connected_clients.append(dict(socket=client_socket, type=DAQ_type))
                    self.settings.child('conn_clients').setValue(self.set_connected_clients_table())
                    self.emit_status(ThreadCommand("Update_Status",
                                                   [DAQ_type + ' connected with ' + address[0] + ':' + str(address[1]),
                                                    'log']))
                    QtWidgets.QApplication.processEvents()

                else:  # Some incoming message from a client
                    # Data received from client, process it
                    try:
                        message = DeSerializer(sock).string_deserialization()
                        if message in ['Done', 'Info', 'Infos', 'Info_xml', 'position_is', 'move_done']:
                            self.process_cmds(message, command_sock=None)
                        elif message == 'Quit':
                            raise Exception("socket disconnect by user")
                        else:
                            self.process_cmds(message, command_sock=sock)

                    # client disconnected, so remove from socket list
                    except Exception as e:
                        self.remove_client(sock)

            self.processing = False

        except Exception as e:
            self.emit_status(ThreadCommand("Update_Status", [str(e), 'log']))

    def send_command(self, sock: Socket, command="move_at"):
        """
            Send one of the message contained in self.message_list toward a socket with identity socket_type.
            First send the length of the command with 4bytes.

            =============== =========== ==========================
            **Parameters**    **Type**    **Description**
            *sock*             ???        The current socket
            *command*         string      The command as a string
            =============== =========== ==========================

            See Also
            --------
            utility_classes.DAQ_Viewer_base.emit_status, daq_utils.ThreadCommand, message_to_bytes
        """
        if command not in self.message_list:
            self.emit_status(
                ThreadCommand("Update_Status", [f'Command: {command} is not in the specified list: {self.message_list}',
                                                'log']))
            return

        if sock is not None:
            sock.check_sended_with_serializer(command)

    def emit_status(self, status):
        print(status)

    def read_data(self, sock):
        pass

    def send_data(self, sock, data):
        pass

    def command_done(self, command_sock):
        pass

    def command_to_from_client(self, command):
        pass

    def process_cmds(self, command, command_sock=None):
        """
            Process the given command.
        """
        if command not in self.message_list:
            self.emit_status(
                ThreadCommand("Update_Status", [f'Command: {command} is not in the specified list: {self.message_list}',
                                                'log']))
            return

        if command == 'Done':  # means the given socket finished grabbing data and is ready to send them
            self.command_done(command_sock)

        elif command == "Infos":
            """replace entirely the client settings information on the server widget
            should be done as the init of the client module"""
            try:
                sock = self.find_socket_within_connected_clients(self.client_type)
                if sock is not None:  # if client self.client_type is connected then send it the command
                    self.read_infos(sock)

            except Exception as e:
                self.emit_status(ThreadCommand("Update_Status", [str(e), 'log']))

        elif command == 'Info_xml':
            """update the state of one of the client settings on the server widget"""
            sock = self.find_socket_within_connected_clients(self.client_type)
            if sock is not None:
                self.read_info_xml(sock)

        elif command == "Info":
            # add a custom info (as a string value) in the server widget settings. To be used if the client is not a
            # PyMoDAQ's module

            try:
                sock = self.find_socket_within_connected_clients(self.client_type)
                if sock is not None:  # if client self.client_type is connected then send it the command
                    self.read_info(sock)
            except Exception as e:
                self.emit_status(ThreadCommand("Update_Status", [str(e), 'log']))

        else:
            self.command_to_from_client(command)

    def read_infos(self, sock: Socket = None, infos=''):
        if sock is not None:
            infos = DeSerializer(sock).string_deserialization()
        params = ioxml.XML_string_to_parameter(infos)

        param_state = {'title': 'Infos Client:', 'name': 'settings_client', 'type': 'group', 'children': params}
        self.settings.child('settings_client').restoreState(param_state)

    def read_info_xml(self, sock: Socket, path=['settings', 'apathinsettings'], param_xml=''):
        if sock is not None:
            deser = DeSerializer(sock)
            path = deser.list_deserialization()
            param_xml = deser.string_deserialization()
        try:
            param_dict = ioxml.XML_string_to_parameter(param_xml)[0]
        except Exception as e:
            raise Exception(f'Invalid xml structure for TCP server settings: {str(e)}')
        try:
            param_here = self.settings.child('settings_client', *path[1:])
        except Exception as e:
            raise Exception(f'Invalid path for TCP server settings: {str(e)}')
        param_here.restoreState(param_dict)

    def read_info(self, sock: Socket=None, test_info='an_info', test_value=''):
        """
        if the client is not from PyMoDAQ it can use this method to display some info into the server widget
        """
        # #first get the info type
        if sock is None:
            info = test_info
            data = test_value
        else:
            deser = DeSerializer(sock)
            info = deser.string_deserialization()
            data = deser.string_deserialization()

        if info not in putils.iter_children(self.settings.child('infos'), []):
            self.settings.child('infos').addChild({'name': info, 'type': 'str', 'value': data})
            pass
        else:
            self.settings.child('infos', info).setValue(data)


class MockServer(TCPServer):
    params = []

    def __init__(self, client_type='GRABBER'):
        super().__init__(client_type)

        self.settings = Parameter.create(name='settings', type='group', children=tcp_parameters)


class MockDataGrabber:

    def __init__(self, grabber_dim='2D'):
        super().__init__()
        self.Nx = 256
        self.Ny = 128
        self.x_axis = np.linspace(0, self.Nx-1, self.Nx)
        self.y_axis = np.linspace(0, self.Ny-1, self.Ny)
        self.grabber_dim = grabber_dim

    def grab(self) -> List[DataFromPlugins]:
        if self.grabber_dim == '0D':
            return [DataFromPlugins(data=[np.array([np.random.rand()])])]

        elif self.grabber_dim == '1D':
            return [DataFromPlugins(data=[mutils.gauss1D(self.x_axis, 128, 25) +
                                          np.random.rand(self.Nx)])]

        elif self.grabber_dim == '2D':
            return [DataFromPlugins(data=[mutils.gauss2D(self.x_axis, 128, 65,
                                                         self.y_axis, 60, 10) +
                                          np.random.rand(self.Ny, self.Nx)])]


class Grabber(QObject):
    command_tcpip = Signal(ThreadCommand)

    def __init__(self, grab_method=None):
        super().__init__()
        self.send_to_tcpip = False
        self.grab_method = grab_method

    def connect_tcp_ip(self, ip='localhost', port=6341):
        self.tcpclient_thread = QThread()
        tcpclient = TCPClient(ip, port=port)

        tcpclient.moveToThread(self.tcpclient_thread)
        self.tcpclient_thread.tcpclient = tcpclient
        tcpclient.cmd_signal.connect(self.process_tcpip_cmds)

        self.command_tcpip[ThreadCommand].connect(tcpclient.queue_command)

        self.tcpclient_thread.start()
        #tcpclient.init_connection(extra_commands=[ThreadCommand('get_axis', )])
        tcpclient.init_connection()
        self.send_to_tcpip = True

    def snapshot(self, info='', send_to_tcpip=True):
        self.grab_data()

    def grab_data(self):
        """
            Do a grab session using 2 profile :
                * if grab pb checked do  a continous save and send an "update_channels" thread command and a "grab" too.
                * if not send a "stop_grab" thread command with settings "main settings-naverage" node value as an attribute.

            See Also
            --------
            daq_utils.ThreadCommand, set_enabled_Ini_buttons
        """
        data = self.grab_method()
        self.command_tcpip.emit(ThreadCommand('data_ready', data))

    @Slot(ThreadCommand)
    def process_tcpip_cmds(self, status):
        if 'Send Data' in status.command:
            self.snapshot('', send_to_tcpip=True)

        elif status.command == 'connected':
            #self.settings.child('main_settings', 'tcpip', 'tcp_connected').setValue(True)
            pass

        elif status.command == 'disconnected':
            #self.settings.child('main_settings', 'tcpip', 'tcp_connected').setValue(False)
            pass

        elif status.command == 'Update_Status':
            print(status)


if __name__ == '__main__':  # pragma: no cover
    import sys
    app = QtWidgets.QApplication(sys.argv)
    mockdata_grabber = MockDataGrabber('2D')
    grabber = Grabber(mockdata_grabber.grab)
    grabber.connect_tcp_ip()

    sys.exit(app.exec_())
